"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.intern.SpecificProjection import SpecificProjection
from ANNarchy.intern.ConfigManagement import ConfigManager, _check_paradigm
from ANNarchy.intern import Messages

from ANNarchy.generator.Utils import tabify, remove_trailing_spaces

from copy import deepcopy

from .PoolingTemplate import *
from .Utils import SharedSynapse

# Indices used for each dimension
indices = ['i', 'j', 'k', 'l', 'm', 'n']

class Pooling(SpecificProjection):
    """
    Performs a pooling operation (e.g. max.pooling) on the pre-synaptic population.

    Each post-synaptic neuron covers a specific region (``extent``) of the pre-synaptic
    population, over which the result of the operation on firing rates will be
    assigned to sum(target).

    The extent is automatically computed using the geometry of the populations, but can be specified in the `pooling()`` methods.

    Example:

    ```python
    import ANNarchy as ann
    from ANNarchy.extensions.convolution import Pooling

    net = ann.Network()
    inp = net.create(geometry=(100, 100), neuron=ann.Neuron(parameters="r = 0.0"))
    pop = net.create(geometry=(50, 50), neuron=ann.Neuron(equations="r = sum(exc)"))
    
    proj = net.connect(Pooling(inp, pop, 'exc', operation='max')) # max-pooling
    proj.pooling() # extent=(2, 2) is implicit
    ```

    :param pre: pre-synaptic population (either its name or a ``Population`` object).
    :param post: post-synaptic population (either its name or a ``Population`` object).
    :param target: type of the connection
    :param operation: pooling function to be applied ("max", "min", "mean")
    """
    def __init__(self, pre, post, target, psp="pre.r", operation="max", name=None, copied=False, net_id=0):

        # Sanity check
        if not operation in ["max", "mean", "min"]:
            Messages._error("Pooling: the operation must be either 'max', 'mean' or 'min'.")
        self.operation = operation

        # Store for _copy
        self.psp = psp

        SpecificProjection.__init__(
            self,
            pre,
            post,
            target,
            synapse=SharedSynapse(
                psp=psp, 
                operation=operation, 
                name="Pooling with '"+operation+"' operation",
                description=operation+"-pooling operation over the pre-synaptic population."
            ),
            name=name,
            copied=copied,
            net_id=net_id
        )

        # check dimensions of populations, should not exceed 4
        self.dim_pre = self.pre.dimension
        self.dim_post = self.post.dimension
        if self.dim_post > 4:
            Messages._error('Pooling: Too many dimensions for the post-synaptic population (maximum 4).')
        if self.dim_pre > 4:
            Messages._error('Pooling: Too many dimensions for the pre-synaptic population (maximum 4).')

        # Disable saving
        self._saveable = False

        # Flags set by connect_xxx() but needed by _copy()
        self.extent = None
        self.delays = None
        self._connection_method = None
        self._connection_args = None
        self._connection_delay = None
        self._storage_format = None



    def connect_pooling(self, extent:tuple=None, delays:float=0.0):
        return self.pooling(extent, delays)

    def pooling(self, extent:tuple=None, delays:float=0.0):
        """
        Connects the `Pooling` projection.

        :param extent: extent of the pooling area expressed in the geometry of the pre-synaptic population (e.g ``(2, 2)``). In each dimension, the product of this extent with the number of neurons in the post-synaptic population must be equal to the number of pre-synaptic neurons. Default: None.
        :param delays: synaptic delay in ms
        """

        # process extent
        self.extent_init = extent
        if extent is None:  # compute the extent automatically
            if self.pre.dimension != self.post.dimension:
                Messages._error(
                    'Pooling: If you do not provide the extent parameter, the two populations must have the same number of dimensions.')

            extent = list(self.pre.geometry)
            for dim in range(self.pre.dimension):
                extent[dim] /= self.post.geometry[dim]
                if self.pre.geometry[dim] != extent[dim] * self.post.geometry[dim]:
                    Messages._error(
                        'Pooling: Unable to compute the extent of the pooling area: the number of neurons do not match.')

        elif not isinstance(extent, tuple):
            Messages._error('Pooling: You must provide a tuple for the extent of the pooling operation.')

        self.extent = list(extent)
        if len(self.extent) < self.pre.dimension:
            Messages._error('Pooling: You must provide a tuple for the extent of the pooling operation.')

        # Process the delays
        self.delays = float(delays)
        if not isinstance(delays, (int, float)):
            Messages._error('Convolutions can only have constant delays.')

        # Generate the pre-synaptic coordinates
        self._generate_extent_coordinates()

        # create fake LIL
        self._create()

        return self

    def _copy(self, pre, post, net_id=None):
        "Returns a copy of the projection when creating networks.  Internal use only."
        copied_proj = Pooling(
            pre=pre, post=post, target=self.target, psp=self.psp,
            operation=self.operation, name=self.name, copied=True,
            net_id = self.net_id if not net_id else net_id
        )

        copied_proj.extent = self.extent
        copied_proj.delays = self.delays

        if self.extent is not None:
            copied_proj._generate_extent_coordinates()
            copied_proj._create()

        copied_proj._connection_method = self._connection_method
        copied_proj._connection_args = self._connection_args
        copied_proj._connection_delay = self._connection_delay
        copied_proj._storage_format = self._storage_format
        return copied_proj

    def _create(self) -> None:
        """
        create fake LIL object, just for compilation process
        """
        try:
            from ANNarchy.cython_ext.Connector import LILConnectivity
        except Exception as e:
            Messages._print(e)
            Messages._error('ANNarchy was not successfully installed.')

        lil = LILConnectivity()
        lil.max_delay = self.delays
        lil.uniform_delay = self.delays
        self.connector_name = "Pooling"
        self.connector_description = "Pooling"
        self._store_connectivity(self._load_from_lil, (lil, ), self.delays, storage_format="lil", storage_order="post_to_pre")

    def _connect(self, module):
        """
        Builds up dendrites either from list or dictionary. Called by instantiate().
        """
        if not self._connection_method:
            Messages._error(
                'Pooling: The projection between ' + self.pre.name + ' and ' + self.post.name + ' is declared but not connected.')

        # Create the Cython instance
        self.cyInstance = getattr(module, 'proj' + str(self.id) + '_wrapper')()
        self.cyInstance.pre_coords = self.pre_coordinates

        return True

    def _generate_extent_coordinates(self) -> list[list[int]]:
        """
        Generates for each post-neuron the position of the top-left corner, where the pooling should be applied.
        """
        # Generates coordinates TODO: Find a more robust way!
        coords = [[] for i in range(self.post.size)]
        if self.dim_pre == 1:
            rk = 0
            for i in range(self.post.geometry[0]):
                coords[rk] = int([i * self.extent[0]])
                rk += 1
        elif self.dim_pre == 2:
            rk = 0
            for i in range(self.post.geometry[0]):
                if self.dim_post > 1:
                    for j in range(self.post.geometry[1]):
                        coords[rk] = [int(i * self.extent[0]), int(j * self.extent[1])]
                        rk += 1
                else: # over the whole second axis
                    coords[rk] = [i * self.extent[0], 0]
                    rk += 1

        elif self.dim_pre == 3:
            rk = 0
            for i in range(self.post.geometry[0]):
                for j in range(self.post.geometry[1]):
                    if self.dim_post > 2:
                        for k in range(self.post.geometry[2]):
                            coords[rk] = [int(i * self.extent[0]), int(j * self.extent[1]), int(k * self.extent[2])]
                            rk += 1
                    else: # over the whole third axis
                        coords[rk] = [int(i * self.extent[0]), int(j * self.extent[1]), int(0)]
                        rk += 1

        elif self.dim_pre == 4: # TODO: post has less than 4 dimensions
            rk = 0
            for i in range(self.post.geometry[0]):
                for j in range(self.post.geometry[1]):
                    for k in range(self.post.geometry[2]):
                        for l in range(self.post.geometry[3]):
                            coords[rk] = [int(i * self.extent[0]), int(j * self.extent[1]), int(k * self.extent[2]), int(l * self.extent[3])]
                            rk += 1
        # Save the result
        self.pre_coordinates = coords

    def _generate(self):
        """
        Overrides the default code generation.
        """
        # Convolve_code
        convolve_code, sum_code = self._generate_pooling_code()

        # Generate the code
        if _check_paradigm("openmp", self.net_id):
            self._generate_omp(convolve_code, sum_code)
        elif _check_paradigm("cuda", self.net_id):
            self._generate_cuda()
        else:
            Messages._error("Pooling: not implemented for the configured paradigm")

    def _generate_pooling_code(self):
        """
        Generate loop statements for the desired pooling operation.
        """
        # Operation to be performed: sum, max, min, mean
        operation = self.synapse_type.operation

        # Main code
        # default value for sum in code depends on operation
        sum_default = "0.0"
        if self.synapse_type.operation == "min":
            sum_default = "std::numeric_limits<%(float_prec)s>::max()" % {'float_prec': ConfigManager().get('precision', self.net_id)}
        elif self.synapse_type.operation == "max":
            sum_default = "std::numeric_limits<%(float_prec)s>::min()" % {'float_prec': ConfigManager().get('precision', self.net_id)}

        code = """
            sum = %(sum_default)s;
""" % {'sum_default': sum_default}

        # Generate for loops
        for dim in range(self.dim_pre):
            ind_dict = {
                'index': indices[dim],
                'size': self.extent[dim]
            }
            if self.extent[dim] > 1:
                code += """
            for(int %(index)s_w = 0; %(index)s_w < %(size)s; %(index)s_w++){
    """ % ind_dict

        # Compute indices
        for dim in range(self.dim_pre):
            ind_dict = {
                'index': indices[dim],
                'dim': dim
            }
            if self.extent[dim] > 1:
                code += """
                int %(index)s_pre = coord[%(dim)s] + %(index)s_w;""" % ind_dict
            else:
                code += """
                int %(index)s_pre = coord[%(dim)s];""" % ind_dict

        # Check indices
        for dim in range(self.dim_pre):
            ind_dict = {
                'index': indices[dim],
                'max_size': self.pre.geometry[dim] - 1
            }
            code += """
                if ((%(index)s_pre < 0) ||(%(index)s_pre > %(max_size)s)){
                    continue;
                }""" % ind_dict

        # Compute pre-synaptic rank
        code += """
                rk_pre = %(value)s;""" % {'value': self._coordinates_to_rank('pre', self.pre.geometry)}

        # Compute the value to pool
        psp = self.synapse_type.description['psp']['cpp'] % {
            'id_pre': self.pre.id,
            'id_post': self.post.id,
            'local_index': '[i][j]',
            'global_index': '[i]',
            'pre_index': '[rk_pre]',
            'post_index': '[rk_post]',
            'pre_prefix': 'pop'+str(self.pre.id)+'->',
            'post_prefix': 'pop'+str(self.post.id)+'->'
        }

        # Delays
        if self.delays > ConfigManager().get('dt', self.net_id):
            psp = psp.replace(
                'pop%(id_pre)s->r[rk_pre]' % {'id_pre': self.pre.id},
                'pop%(id_pre)s->_delayed_r[%(delay)s][rk_pre]' % {'id_pre': self.pre.id, 'delay': str(int(self.delays/ConfigManager().get('dt', self.net_id))-1)}
            )

        # Apply the operation
        if operation == "max":
            code += """
                %(float_prec)s _psp = %(psp)s;
                if(_psp > sum) sum = _psp;"""
        elif operation == "min":
            code += """
                %(float_prec)s _psp = %(psp)s;
                if(_psp < sum) sum = _psp;"""
        elif operation == "sum":
            code += """
                sum += %(psp)s;"""
        elif operation == "mean":
            code += """
                sum += %(psp)s;"""
        else:
            Messages._error('SharedProjection: Operation', operation, 'is not implemented yet for shared projections with pooling.')

        # Close for loops
        for dim in range(self.dim_pre):
            if self.extent[dim] > 1:
                code += """
            }"""

        impl_code = code % {
            'id_proj': self.id,
            'target': self.target,
            'id_pre': self.pre.id, 'name_pre': self.pre.name, 'size_pre': self.pre.size,
            'id_post': self.post.id, 'name_post': self.post.name, 'size_post': self.post.size,
            'psp': psp,
            'float_prec': ConfigManager().get('precision', self.net_id)
        }

        if operation == "mean":
            size = 1
            for dim in range(self.pre.dimension):
                size *= self.extent[dim]
            sum_code = "sum / " + str(size)
        else:
            sum_code = "sum"

        return impl_code, sum_code

    def _generate_omp(self, convolve_code, sum_code):
        """
        Update the ProjectionGenerator._specific_template structure and bypass the standard openMP code generation.

        :param convolve_code:
        :param sum_code:
        """

        # Default value for sum in code depends on operation
        sum_default = "0.0"
        if self.synapse_type.operation == "min":
            sum_default = "std::numeric_limits<%(float_prec)s>::max()" % {'float_prec': ConfigManager().get('precision', self.net_id)}
        elif self.synapse_type.operation == "max":
            sum_default = "std::numeric_limits<%(float_prec)s>::min()" % {'float_prec': ConfigManager().get('precision', self.net_id)}

        # Specific template for generation
        pool_dict = deepcopy(pooling_template_omp)
        for key, value in pool_dict.items():
            value = value % {
                'id_proj': self.id,
                'size_post': self.post.size,
                'sum_default': sum_default,
                'float_prec': ConfigManager().get('precision', self.net_id)
            }
            pool_dict[key] = value
        self._specific_template.update(pool_dict)

        # OMP code
        omp_code = ""
        if ConfigManager().get('num_threads', self.net_id) > 1:
            omp_code = """
        #pragma omp for private(sum, rk_pre, coord) %(psp_schedule)s""" % {
                'psp_schedule': "" if not 'psp_schedule' in self._omp_config.keys() else self._omp_config[
                    'psp_schedule']}

        # HD ( 16.10.2015 ):
        # pre-load delayed firing rate in a local array, so we
        # prevent multiple accesses to pop%(id_pre)s->_delayed_r[%(delay)s]
        if self.delays > ConfigManager().get('dt', self.net_id):
            pre_load_r = """
        // pre-load delayed firing rate
        auto delayed_r = pop%(id_pre)s->_delayed_r[%(delay)s];
        """ % {'id_pre': self.pre.id, 'delay': str(int(self.delays / ConfigManager().get('dt', self.net_id)) - 1)}
        else:
            pre_load_r = ""

        # Target variable depends on neuron type
        target_code = "_sum_%(target)s" if self.post.neuron_type.type=="rate" else "g_%(target)s"
        target_code %= {'target': self.target}

        # Compute sum
        wsum = """
        if ( _transmission && pop%(id_pre)s->_active ) {
        std::vector<int> coord;
""" + pre_load_r + """
        %(omp_code)s
        for(int i = 0; i < %(size_post)s; i++){
            coord = pre_coords[i];
""" + convolve_code + """
            pop%(id_post)s->%(target)s[i] += """ + sum_code + """;
        } // for
        } // if
"""


        # Dictionary keys
        psp_dict = {
            'id_proj': self.id,
            'target': target_code,
            'id_pre': self.pre.id, 'name_pre': self.pre.name,
            'size_pre': self.pre.size,
            'id_post': self.post.id, 'name_post': self.post.name,
            'size_post': self.post.size,
            'omp_code': omp_code,
            'convolve_code': convolve_code
        }

        # Psp code
        self._specific_template['psp_code'] = wsum % psp_dict
        self._specific_template['size_in_bytes'] = """
        // connectivity
        size_in_bytes += sizeof(std::vector<int>);
        size_in_bytes += pre_coords.capacity() * sizeof(int);

        size_in_bytes += sizeof(std::vector<std::vector<int>>);
        size_in_bytes += pre_coords.capacity() * sizeof(std::vector<int>);
        for (auto it = pre_coords.begin(); it != pre_coords.end(); it++) {
            size_in_bytes += it->capacity() * sizeof(int);
        }
"""

        # Clean-up
        self._specific_template['clear_container'] = pooling_template_omp["clear"]

    def _generate_cuda(self):
        """
        Update the ProjectionGenerator._specific_template structure and bypass the standard CUDA code generation.
        """
        # Extract operation and pre-synaptic variable which should be processed
        pool_operation = self.synapse_type.operation
        pre_var = self.synapse_type.psp.split(".")[1]

        # default value for sum in code depends on operation
        sum_default = "0.0"
        if pool_operation == "min":
            sum_default = "FLT_MAX"
        elif pool_operation == "max":
            sum_default = "FLT_MIN"

        # operation to perform
        pool_op_code = cuda_op_code[pool_operation] % {'float_prec': ConfigManager().get('precision', self.net_id)}

        # mean operation requires one additional computation
        if pool_operation == "mean":
            size = 1
            for dim in range(self.pre.dimension):
                size *= self.extent[dim]
            final_result = "psp[bIdx] += local_res/" + str(size) + ";"
        else:
            final_result = "psp[bIdx] += local_res;"

        # result dictionary with code for
        # body, call and header
        pool_template = {}
        base_ids = {
            'id_proj': self.id,
            'id_pre': self.pre.id,
            'id_post': self.post.id,
            'target': self.target,
            'float_prec': ConfigManager().get('precision', self.net_id),
            'size_post': self.post.size # TODO: population views?
        }

        # The correct templates depends on both
        # kernel-geometry and extent
        if len(self.pre.geometry) == 2:
            # For small extents, we compute multiple coords within one warp. If one extent can fill alone
            # a half-warp we switch to the other implementation.
            if self.extent[0] < 6:

                pool_op_reduce_code = cuda_pooling_code_2d_small_extent['reduce_code'][pool_operation] % {
                    'float_prec': ConfigManager().get('precision', self.net_id),
                    'row_extent': int(self.extent[0]),
                    'col_extent': int(self.extent[1])
                }

                pool_dict = deepcopy(base_ids)
                pool_dict.update({
                    'sum_default': sum_default,
                    'row_extent': int(self.extent[0]),
                    'col_extent': int(self.extent[1]),
                    'row_size': int(self.pre.geometry[0]),
                    'col_size': int(self.pre.geometry[1]),
                    'operation': tabify(pool_op_code, 3),
                    'operation_reduce': pool_op_reduce_code,
                    'final_result': final_result
                })

                pool_template['psp_body'] = cuda_pooling_code_2d_small_extent['psp_body'] % pool_dict
                pool_template['psp_invoke'] = cuda_pooling_code_2d_small_extent['psp_invoke'] % pool_dict
                pool_template['psp_header'] = cuda_pooling_code_2d_small_extent['psp_header'] % pool_dict
                pool_template['psp_call'] = cuda_pooling_code_2d_small_extent['psp_call'] % pool_dict

            else:
                pool_op_reduce_code = cuda_pooling_code_2d['reduce_code'][pool_operation] % {
                    'float_prec': ConfigManager().get('precision', self.net_id),
                    'row_extent': int(self.extent[0]),
                    'col_extent': int(self.extent[1])
                }

                pool_dict = deepcopy(base_ids)
                pool_dict.update({
                    'sum_default': sum_default,
                    'row_extent': int(self.extent[0]),
                    'col_extent': int(self.extent[1]),
                    'row_size': int(self.pre.geometry[0]),
                    'col_size': int(self.pre.geometry[1]),
                    'operation': tabify(pool_op_code, 3),
                    'operation_reduce': tabify(pool_op_reduce_code, 2),
                    'final_result': final_result
                })

                pool_template['psp_body'] = remove_trailing_spaces(cuda_pooling_code_2d['psp_body'] % pool_dict)
                pool_template['psp_invoke'] = remove_trailing_spaces(cuda_pooling_code_2d['psp_invoke'] % pool_dict)
                pool_template['psp_header'] = cuda_pooling_code_2d['psp_header'] % pool_dict
                pool_template['psp_call'] = cuda_pooling_code_2d['psp_call'] % pool_dict

        elif len(self.pre.geometry) == 3:

            pool_dict = deepcopy(base_ids)
            pool_dict.update({
                'sum_default': sum_default,
                'row_extent': self.extent[0],
                'col_extent': self.extent[1],
                'plane_extent': self.extent[2],
                'row_size': self.pre.geometry[0],
                'col_size': self.pre.geometry[1],
                'plane_size': self.pre.geometry[2],
                'pre_var': pre_var,
                'operation': tabify(pool_op_code, 4),
                'final_result': final_result
            })

            pool_template['psp_body'] = remove_trailing_spaces(cuda_pooling_code_3d['psp_body'] % pool_dict)
            pool_template['psp_invoke'] = remove_trailing_spaces(cuda_pooling_code_3d['psp_invoke'] % pool_dict)
            pool_template['psp_header'] = cuda_pooling_code_3d['psp_header'] % pool_dict
            pool_template['psp_call'] = cuda_pooling_code_3d['psp_call'] % pool_dict

        else:
            raise NotImplementedError

        # Post-neuron is a spike neuron (e.g., part of ANN-to-SNN conversion)
        if self.post.neuron_type.type == "spike":
            pool_template['psp_call'] = pool_template['psp_call'].replace("gpu__sum_"+self.target, "gpu_g_"+self.target)

        # Update psp fields
        self._specific_template.update(pool_template)

        # Specific template for generation (wrapper, etc)
        pool_dict = deepcopy(pooling_template_cuda)
        for key, value in pool_dict.items():
            value = value % base_ids
            pool_dict[key] = value
        self._specific_template.update(pool_dict)

        self._specific_template['wrapper_connector_call'] = ""
        self._specific_template['access_parameters_variables'] = ""

        self._specific_template['size_in_bytes'] = "//TODO:\n"

        # Clean-up
        self._specific_template['clear_container'] = pooling_template_cuda["clear"]

    @staticmethod
    def _coordinates_to_rank(name, geometry):
        """
        Generate the code for array access, for instance used
        for pre-synaptic ranks.
        """
        dim = len(geometry)

        txt = ""

        for d in range(dim):
            if txt == "":   # first coordinate is special
                txt = indices[0] + "_" + name
            else:
                txt = str(geometry[d]) + '*(' + txt + ') + ' + indices[d] + '_' + name

        return txt

    ##############################
    ## Override useless methods
    ##############################
    def _data(self):
        "Disable saving."
        desc = {}
        desc['post_ranks'] = self.post_ranks
        desc['attributes'] = self.attributes
        desc['parameters'] = self.parameters
        desc['variables'] = self.variables

        desc['dendrites'] = []
        desc['number_of_synapses'] = 0
        return desc

    def save_connectivity(self, filename):
        "Not available."
        Messages._warning('Pooling projections can not be saved.')
    def save(self, filename):
        "Not available."
        Messages._warning('Pooling projections can not be saved.')
    def load(self, filename):
        "Not available."
        Messages._warning('Pooling projections can not be loaded.')
    def receptive_fields(self, variable = 'w', in_post_geometry = True):
        "Not available."
        Messages._warning('Pooling projections can not display receptive fields.')
    def connectivity_matrix(self, fill=0.0):
        "Not available."
        Messages._warning('Pooling projections can not display connectivity matrices.')



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

# Code which should be added prior to kernels
# (directly imported by CodeGenerator if needed)
additional_global_functions = ""

launch_config = {
    'init': """
        _threads_per_block = 0;
        _nb_blocks = 0;

    #ifdef _DEBUG
        std::cout << "Kernel configuration is a fixed 2D kernel" << std::endl;
    #endif
""",
    'update': """
        // Generate the kernel launch configuration
        _threads_per_block = 0;
        _nb_blocks = 0;

    #ifdef _DEBUG
        std::cout << "Kernel configuration is a fixed 2D kernel" << std::endl;
    #endif
"""
}

attribute_decl = {
    'local': """
    // Local %(attr_type)s %(name)s
    std::vector< %(type)s > %(name)s;
    %(type)s* gpu_%(name)s;
    long int %(name)s_device_to_host;
    bool %(name)s_host_to_device;
""",
    'semiglobal': """
    // Semiglobal %(attr_type)s %(name)s
    std::vector< %(type)s >  %(name)s ;
    %(type)s* gpu_%(name)s;
    long int %(name)s_device_to_host;
    bool %(name)s_host_to_device;
""",
    'global': {
        'parameter': """
    // Global %(attr_type)s %(name)s
    %(type)s %(name)s;
""",
        'variable': """
    // Global %(attr_type)s %(name)s
    %(type)s %(name)s;
    %(type)s* gpu_%(name)s;
    long int %(name)s_device_to_host;
    bool %(name)s_host_to_device;
"""
    }
}


attribute_cpp_init = {
    'local':
"""
        // Local %(attr_type)s %(name)s
        %(name)s = init_matrix_variable<%(type)s>(static_cast<%(type)s>(%(init)s));
        gpu_%(name)s = init_matrix_variable_gpu<%(type)s>(%(name)s);
        %(name)s_host_to_device = true;
        %(name)s_device_to_host = t;
""",
    'semiglobal':
"""
        // Semiglobal %(attr_type)s %(name)s
        %(name)s = init_vector_variable<%(type)s>(static_cast<%(type)s>(%(init)s));
        gpu_%(name)s = init_vector_variable_gpu<%(type)s>(%(name)s);
        %(name)s_host_to_device = true;
        %(name)s_device_to_host = t;
""",
    'global': {
        'parameter': """
        // Global %(attr_type)s %(name)s
        %(name)s = 0.0;
    """,
        'variable': """
        // Global %(attr_type)s %(name)s
        %(name)s = static_cast<%(type)s>(%(init)s);
        cudaMalloc((void**)&gpu_%(name)s, sizeof(%(type)s));
        %(name)s_host_to_device = true;
        %(name)s_device_to_host = t;
"""
    }
}

attribute_cpp_size = {
    'local': """
        // Local %(attr_type)s %(name)s
        size_in_bytes += sizeof(std::vector<%(ctype)s>);
        size_in_bytes += sizeof(%(ctype)s) * %(name)s.capacity();
""",
    'semiglobal': """
        // Semiglobal %(attr_type)s %(name)s
        size_in_bytes += sizeof(std::vector<%(ctype)s>);
        size_in_bytes += sizeof(%(ctype)s) * %(name)s.capacity();
""",
    'global': """
        // Global %(attr_type)s %(name)s
        size_in_bytes += sizeof(%(ctype)s);
"""
}

attribute_cpp_delete = {
    'local': """
        // %(name)s - host
        %(name)s.clear();
        %(name)s.shrink_to_fit();

        // %(name)s - device
        cudaFree(gpu_%(name)s);
""",
    'semiglobal': """
        // %(name)s - host
        %(name)s.clear();
        %(name)s.shrink_to_fit();

        // %(name)s - device
        cudaFree(gpu_%(name)s);
""",
    'global': ""
}

attribute_host_to_device = {
    'local': """
        // %(name)s: local
        if ( %(name)s_host_to_device )
        {
        #ifdef _DEBUG
            std::cout << "HtoD: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( gpu_%(name)s, %(name)s.data(), this->num_rows_ * this->num_columns_ * sizeof( %(type)s ), cudaMemcpyHostToDevice);
            %(name)s_host_to_device = false;
        #ifdef _DEBUG
            cudaError_t err = cudaGetLastError();
            if ( err!= cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err) << std::endl;
        #endif
        }
""",
    'semiglobal': """
        // %(name)s: semiglobal
        if ( %(name)s_host_to_device )
        {
        #ifdef _DEBUG
            std::cout << "HtoD: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( gpu_%(name)s, %(name)s.data(), this->num_rows_ * sizeof( %(type)s ), cudaMemcpyHostToDevice);
            %(name)s_host_to_device = false;
        #ifdef _DEBUG
            cudaError_t err = cudaGetLastError();
            if ( err!= cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err) << std::endl;
        #endif
        }
""",
    'global': """
        // %(name)s: global
        if ( %(name)s_host_to_device )
        {
        #ifdef _DEBUG
            std::cout << "HtoD: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            //cudaMemcpy( gpu_%(name)s, &%(name)s, sizeof( %(type)s ), cudaMemcpyHostToDevice);
            %(name)s_host_to_device = false;
        #ifdef _DEBUG
            cudaError_t err = cudaGetLastError();
            if ( err!= cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err) << std::endl;
        #endif
        }
"""
}

attribute_device_to_host = {
    'local': """
        // %(name)s: local
        if ( %(name)s_device_to_host < t ) {
        #ifdef _DEBUG
            std::cout << "DtoH: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( %(name)s.data(), gpu_%(name)s, num_rows_ * num_columns_ * sizeof( %(type)s ), cudaMemcpyDeviceToHost);
        #ifdef _DEBUG
            cudaError_t err_%(name)s = cudaGetLastError();
            if ( err_%(name)s != cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err_%(name)s) << std::endl;
        #endif
            %(name)s_device_to_host = t;
        }
""",
    'semiglobal': """
            // %(name)s: semiglobal
        #ifdef _DEBUG
            std::cout << "DtoH: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            //cudaMemcpy( %(name)s.data(), gpu_%(name)s, post_ranks_.size() * sizeof(%(type)s), cudaMemcpyDeviceToHost);
        #ifdef _DEBUG
            cudaError_t err_%(name)s = cudaGetLastError();
            if ( err_%(name)s != cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err_%(name)s) << std::endl;
        #endif
""",
    'global': """
            // %(name)s: global
        #ifdef _DEBUG
            std::cout << "DtoH: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            //cudaMemcpy( &%(name)s, gpu_%(name)s, sizeof(%(type)s), cudaMemcpyDeviceToHost);
        #ifdef _DEBUG
            cudaError_t err_%(name)s = cudaGetLastError();
            if ( err_%(name)s != cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err_%(name)s) << std::endl;
        #endif
"""
}

delay = {
    # A single value for all synapses
    'uniform': {
        'declare': """
    // Uniform delay
    int delay;

    int get_delay() { return delay; }
    int get_dendrite_delay(int idx) { return delay; }
    void set_delay(int delay) { this->delay = delay; }
""",
        'init': """
    delay = delays[0][0];
"""
    },
    # An individual value for each synapse
    'nonuniform_rate_coded': None,
    # An individual value for each synapse and a
    # buffer for spike events
    'nonuniform_spiking': None
}

event_driven = {
    'declare': """
    std::vector< long > _last_event;
    long* _gpu_last_event;
""",
    'cpp_init': """
    _last_event = init_matrix_variable<long>(-10000);
    _gpu_last_event = init_matrix_variable_gpu<long>(_last_event);
""",
    'pyx_struct': """
        vector[long] _last_event
""",
    'pyx_wrapper_init': """
        proj%(id_proj)s._last_event = vector[long]( syn._matrix.num_elements(), -10000)
"""
}

#
# Implement the continuous signal transmission for rate-coded synapses.
#
rate_psp_kernel = {
    # Comment to if (tid < 32) block:
    #
    # now that we are using warp-synchronous programming (below)
    # we need to declare our shared memory volatile so that the compiler
    # doesn't reorder stores to it and induce incorrect behavior.
    'device_kernel': {
        'sum':"""
__global__ void cu_proj%(id_proj)s_psp(%(conn_args)s%(add_args)s, %(float_prec)s* %(target_arg)s ) {
    %(idx_type)s rk_post = blockIdx.y*blockDim.y+threadIdx.y;
    %(float_prec)s __shared__ sdata[16][32];
    unsigned int tid = threadIdx.x;

    while( rk_post < post_size ) {
        sdata[threadIdx.y][threadIdx.x] = 0.0;
        %(size_type)s j = rk_post * pre_size + threadIdx.x;

        for (%(idx_type)s rk_pre = threadIdx.x; rk_pre < pre_size; rk_pre+=blockDim.x, j+= blockDim.x) {
            sdata[threadIdx.y][threadIdx.x] += %(psp)s
        }

        __syncthreads();

        // do reduction in shared mem within one warp
        if (threadIdx.x < 16) {
            volatile %(float_prec)s* data = sdata[threadIdx.y];
            data[tid] += data[tid + 16];
            data[tid] += data[tid +  8];
            data[tid] += data[tid +  4];
            data[tid] += data[tid +  2];
            data[tid] += data[tid +  1];
        }

        if (threadIdx.x == 0) %(target_arg)s%(post_index)s += sdata[threadIdx.y][0];
        __syncthreads();
        rk_post += gridDim.y*blockDim.y;
    }
}
"""
    },
    'invoke_kernel': """
void proj%(id_proj)s_psp(RunConfig cfg, %(conn_args)s%(add_args)s, %(float_prec)s* %(target_arg)s) {
    cu_proj%(id_proj)s_psp<<< cfg.nb, cfg.tpb, cfg.smem_size, cfg.stream >>>(
        /* ranks and offsets */
        %(conn_args_call)s
        /* computation data */
        %(add_args_call)s
        /* result */
        %(target_arg_call)s
    );
}
""",
    'kernel_decl': """void proj%(id_proj)s_psp(RunConfig cfg, %(conn_args)s%(add_args)s, %(float_prec)s* %(target_arg)s );
""",
    'host_call': """
    // proj%(id_proj)s: pop%(id_pre)s -> pop%(id_post)s
    if ( pop%(id_post)s->_active && proj%(id_proj)s->_transmission ) {
        // 2D-Kernel: y number rows, x number columns
        auto num_block_rows = static_cast<%(idx_type)s>(ceil(double(proj%(id_proj)s->num_rows())/16.0));
        auto thread_dim = dim3(32, 16, 1);
        auto block_dim = dim3(1, num_block_rows, 1);

        proj%(id_proj)s_psp(
            RunConfig(block_dim, thread_dim, 0, proj%(id_proj)s->stream),
            /* ranks and offsets */
            %(conn_args)s
            /* computation data */
            %(add_args)s
            /* result */
            %(target_arg)s
        );

    #ifdef _DEBUG
        auto err = cudaGetLastError();
        if ( err != cudaSuccess ) {
            std::cout << "cu_proj%(id_proj)s_psp: " << cudaGetErrorString(err) << std::endl;
        }
    #endif
    }
""",
    'thread_init': {
        'float': {
            'sum': "0.0f",
            'min': "FLT_MAX",
            'max': "FLT_MIN",
            'mean': "0.0f"
        },
        'double': {
            'sum': "0.0",
            'min': "DBL_MAX",
            'max': "DBL_MIN",
            'mean': "0.0"
        }
    }
}

spike_event_transmission = {
    'device_kernel': """// gpu device kernel for projection %(id_proj)s
__global__ void cu_proj%(id_proj)s_psp( const long int t, const %(float_prec)s dt, bool plasticity, int *spiked, unsigned int* num_events, %(conn_args_header)s %(kernel_args_header)s ) {
    int rk_post = blockIdx.x ;
    int tid = threadIdx.x;

    while (tid < num_events[0]) {
        int rk_pre = spiked[tid];               // which neuron spiked
        int j = rk_post * column_size + rk_pre; // row-major indexing

        // event-driven
    %(event_driven)s

        // increase of conductance
    %(psp)s

        // pre-spike statements
    %(pre_event)s

        // proceed to the next spiking neuron
        tid += blockDim.x;
    }
}
""",
    'invoke_kernel': """
void proj%(id_proj)s_psp(RunConfig cfg, const long int t, const %(float_prec)s dt, bool plasticity, int *spiked, unsigned int* num_events, %(conn_args_header)s %(kernel_args_header)s) {
    cu_proj%(id_proj)s_psp<<< cfg.nb, cfg.tpb, cfg.smem_size, cfg.stream >>>(
        t, dt, plasticity,
        /* pre-synaptic events */
        spiked, num_events,
        /* connectivity */
        %(conn_args_invoke)s
        /* kernel config */
        %(kernel_args_invoke)s
    );
}
""",
    'kernel_decl': """void proj%(id_proj)s_psp(RunConfig cfg, const long int t, const %(float_prec)s dt, bool plasticity, int *spiked, unsigned int* num_events, %(conn_args_header)s %(kernel_args_header)s);
""",
    'host_call': """
    if ( pop%(id_pre)s->_active && (pop%(id_pre)s->spike_count > 0) && proj%(id_proj)s->_transmission ) {
        int tpb = 32;
        int nb = proj%(id_proj)s->num_rows();

        // compute psp
        proj%(id_proj)s_psp(
            RunConfig(nb, tpb, 0, proj%(id_proj)s->stream),
            t, dt, proj%(id_proj)s->_plasticity,
            /* pre-synaptic events */
            pop%(id_pre)s->gpu_spiked, pop%(id_pre)s->gpu_spike_count,
            /* connectivity */
            %(conn_args)s
            /* kernel config */
            %(kernel_args)s
        );
    #ifdef _DEBUG
        cudaDeviceSynchronize();
        cudaError_t err_psp_proj%(id_proj)s = cudaGetLastError();
        if( err_psp_proj%(id_proj)s != cudaSuccess) {
            std::cout << "proj%(id_proj)s_psp (" << t << "): " << std::endl;
            std::cout << "   " << cudaGetErrorString(err_psp_proj%(id_proj)s) << std::endl;
        }
    #endif
    }
"""
}

# Update of global synaptic equations, consist of body (annarchyDevice.cu),
# header and call semantic (take place in ANNarchyHost.cu)
global_synapse_update = {
    'device_kernel': """
// gpu device kernel for projection %(id_proj)s
__global__ void cuProj%(id_proj)s_global_step(
    /* default params */
    const long int, const %(float_prec)s dt
    /* additional params */
    %(kernel_args)s,
    /* plasticity enabled */
    bool plasticity )
{
%(pre_loop)s
%(global_eqs)s
}
""",
    'invoke_kernel': """
void proj%(id_proj)s_global_step(RunConfig cfg, const long int t, %(float_prec)s dt %(kernel_args)s, bool plasticity) {
    cuProj%(id_proj)s_global_step<<< cfg.nb, cfg.tpb, cfg.smem_size, cfg.stream>>>(
        /* default args*/
        t, dt
        /* kernel args */
        %(kernel_args_call)s
        /* synaptic plasticity */
        , plasticity
    );
}
""",
    'kernel_decl': """void proj%(id_proj)s_global_step(RunConfig cfg, const long int t, %(float_prec)s dt %(kernel_args)s, bool plasticity);
""",
    'host_call': """
        // global update
        proj%(id_proj)s_global_step(
            RunConfig(1, 1, 0, proj%(id_proj)s->stream),
            /* default args*/
            t, _dt
            /* kernel args */
            %(kernel_args_call)s
            /* synaptic plasticity */
            , proj%(id_proj)s->_plasticity
        );

    #ifdef _DEBUG
        cudaDeviceSynchronize();
        err = cudaGetLastError();
        if ( global_step != cudaSuccess) {
            std::cout << "proj%(id_proj)s_step: " << cudaGetErrorString( err ) << std::endl;
        }
    #endif
"""
}

# Update of semiglobal synaptic equations, consist of body (annarchyDevice.cu),
# header and call semantic (take place in ANNarchyHost.cu)
semiglobal_synapse_update = {
    'device_kernel': """
// gpu device kernel for projection %(id_proj)s
__global__ void cuProj%(id_proj)s_semiglobal_step(
    /* default params */
    const int post_size, const long int t, const %(float_prec)s dt
    /* additional params */
    %(kernel_args)s,
    /* plasticity enabled */
    bool plasticity
) {
    int i = threadIdx.x + blockIdx.x*blockDim.x;

%(pre_loop)s
    while ( i < post_size ) {
%(semiglobal_eqs)s

        i += gridDim.x * blockDim.x;
    }
}
""",
    'invoke_kernel': """
void proj%(id_proj)s_semiglobal_step(RunConfig cfg, const %(idx_type)s post_size, const long int t, %(float_prec)s dt %(kernel_args)s, bool plasticity) {
    cuProj%(id_proj)s_semiglobal_step<<< cfg.nb, cfg.tpb, cfg.smem_size, cfg.stream >>>(
        /* default args*/
        post_size, t, dt
        /* kernel args */
        %(kernel_args_call)s
        /* synaptic plasticity */
        , plasticity
    );
}
""",
    'kernel_decl': """void proj%(id_proj)s_semiglobal_step(RunConfig cfg, const %(idx_type)s post_size, const long int t, %(float_prec)s dt %(kernel_args)s, bool plasticity);
""",
    'host_call': """
        // semiglobal update
        nb_blocks = ceil( %(float_prec)s(proj%(id_proj)s->num_rows()) / 32.0);
        proj%(id_proj)s_semiglobal_step(
            RunConfig(nb_blocks, 32, 0, proj%(id_proj)s->stream),
            /* default args*/
            proj%(id_proj)s->num_rows(), t, _dt
            /* kernel args */
            %(kernel_args_call)s
            /* synaptic plasticity */
            , proj%(id_proj)s->_plasticity
        );

    #ifdef _DEBUG
        cudaDeviceSynchronize();
        err = cudaGetLastError();
        if ( err != cudaSuccess) {
            std::cout << "proj%(id_proj)s_semiglobal_step: " << cudaGetErrorString( err ) << std::endl;
        }
    #endif
"""
}

# Update of local synaptic equations, consist of body (annarchyDevice.cu),
# header and call semantic (take place in ANNarchyHost.cu)
local_synapse_update = {
    'device_kernel': """
// gpu device kernel for projection %(id_proj)s
__global__ void cuProj%(id_proj)s_local_step(
    /* connectivity */
    %(idx_type)s post_size, %(idx_type)s pre_size, char* mask,
    /* default params */
    const long int t, const %(float_prec)s dt
    /* additional params */
    %(kernel_args)s,
    /* plasticity enabled */
    bool plasticity
) {
    
    %(idx_type)s rk_pre = blockIdx.x * blockDim.x + threadIdx.x;
    if (rk_pre < pre_size) {

%(pre_loop)s
        // Updating local variables of projection %(id_proj)s
        for (%(idx_type)s rk_post = 0; rk_post < post_size; rk_post++)
        {
            %(size_type)s j = rk_pre + rk_post * pre_size;
            if (mask[j]) {
%(local_eqs)s
            }

            j += blockDim.x * gridDim.x;
        }
    }
}
""",
    'invoke_kernel': """
void proj%(id_proj)s_local_step(RunConfig cfg, %(idx_type)s post_size, %(idx_type)s pre_size, char* mask, const long int t, const %(float_prec)s dt %(kernel_args)s, bool plasticity) {
    cuProj%(id_proj)s_local_step<<< cfg.nb, cfg.tpb, cfg.smem_size, cfg.stream >>>(
        /* default args*/
        post_size, pre_size, mask, t, dt
        /* kernel args */
        %(kernel_args_call)s
        /* synaptic plasticity */
        , plasticity
    );
}
""",
    'kernel_decl': """void proj%(id_proj)s_local_step(RunConfig cfg, %(idx_type)s post_size, %(idx_type)s pre_size, char* mask, const long int t, const %(float_prec)s dt %(kernel_args)s, bool plasticity);
""",
    'host_call': """
        // local update
    #if defined (__proj%(id_proj)s_%(target)s_tpb__)
        RunConfig proj%(id_proj)s_local_step_cfg = RunConfig(__proj%(id_proj)s_nb__, __proj%(id_proj)s_%(target)s_tpb__, 0, proj%(id_proj)s->stream);
    #else
        proj%(id_proj)s->_threads_per_block = 32;
        proj%(id_proj)s->_nb_blocks = static_cast<int>(ceil ( static_cast<%(float_prec)s>(pop%(id_pre)s->size) / static_cast<%(float_prec)s>(proj%(id_proj)s->_threads_per_block)));
        RunConfig proj%(id_proj)s_local_step_cfg = RunConfig(proj%(id_proj)s->_nb_blocks, proj%(id_proj)s->_threads_per_block, 0, proj%(id_proj)s->stream);
    #endif
        proj%(id_proj)s_local_step(
            proj%(id_proj)s_local_step_cfg,
            /* default args*/
            pop%(id_post)s->size, pop%(id_pre)s->size, proj%(id_proj)s->device_mask(), t, _dt
            /* kernel args */
            %(kernel_args_call)s
            /* synaptic plasticity */
            , proj%(id_proj)s->_plasticity
        );

    #ifdef _DEBUG
        cudaDeviceSynchronize();
        err = cudaGetLastError();
        if ( err != cudaSuccess) {
            std::cout << "proj%(id_proj)s_step: " << cudaGetErrorString( err ) << std::endl;
        }
    #endif
""",
}

# call semantic for global, semiglobal and local kernel
synapse_update_call = """
    // proj%(id_proj)s: pop%(pre)s -> pop%(post)s
    if ( proj%(id_proj)s->_transmission && proj%(id_proj)s->_update && proj%(id_proj)s->_plasticity && ( (t - proj%(id_proj)s->_update_offset)%%proj%(id_proj)s->_update_period == 0L)) {
        %(float_prec)s _dt = dt * proj%(id_proj)s->_update_period;
#ifdef _DEBUG
    cudaError_t err;
#endif
%(global_call)s
int nb_blocks;
%(semiglobal_call)s
%(local_call)s
    }
"""

spike_postevent = {
    'device_kernel': """// Projection %(id_proj)s: post-synaptic events
__global__ void cuProj%(id_proj)s_postevent(
    // default constants
    const long int t, const %(float_prec)s dt, bool plasticity,
    // events
    int* spiked, long int* pre_last_spike,
    // connectivity
    %(conn_args)s,
    // weights and other arguments
    %(float_prec)s* w %(add_args)s )
{
    // each CUDA block computes one row
    int rk_pre  = spiked[blockIdx.x];
    int j = rk_pre * pre_size + threadIdx.x;
    int j_end = (rk_pre+1)*pre_size;

    while ( j < j_end) {
        if (mask[j]) {
    // event-driven
%(event_driven)s

    // post-event
%(post_code)s
        }
        j += blockDim.x;
    }
}
""",
    'invoke_kernel': """
void proj%(id_proj)s_postevent(RunConfig cfg, const long int t, const %(float_prec)s dt, bool plasticity, int* spiked, long int* pre_last_spike, %(conn_args)s, %(float_prec)s* w %(add_args)s ){
    cuProj%(id_proj)s_postevent<<< cfg.nb, cfg.tpb, cfg.smem_size, cfg.stream >>>(
        t, dt, plasticity,
        /* post-spike and pre-spike time points */
        spiked, pre_last_spike,
        /* connectivity */
        %(conn_args_invoke)s
        /* weights */
        , w
        /* other variables */
        %(add_args_invoke)s
    );
}
""",
    'kernel_decl': "void proj%(id_proj)s_postevent(RunConfig cfg, const long int t, const %(float_prec)s dt, bool plasticity, int* spiked, long int* pre_last_spike, %(conn_args)s, %(float_prec)s* w %(add_args)s );",
    'host_call': """
    if ( proj%(id_proj)s->_transmission && pop%(id_post)s->_active && (pop%(id_post)s->spike_count > 0) ) {
    #if defined (__proj%(id_proj)s_%(target)s_nb__)
        int tpb = __proj%(id_proj)s_%(target)s_tpb__;
    #else
        int tpb = 64;
    #endif

        proj%(id_proj)s_postevent(
            RunConfig(pop%(id_post)s->spike_count, tpb, 0, proj%(id_proj)s->stream),
            t, dt, proj%(id_proj)s->_plasticity,
            /* post-spike and pre-spike time points */
            pop%(id_post)s->gpu_spiked, pop%(id_pre)s->gpu_last_spike,
            /* connectivity */
            %(conn_args)s
            /* weights */
            , proj%(id_proj)s->gpu_w
            /* other variables */
            %(add_args)s
        );
    #ifdef _DEBUG
        cudaDeviceSynchronize();
        cudaError_t proj%(id_proj)s_postevent = cudaGetLastError();
        if (proj%(id_proj)s_postevent != cudaSuccess) {
            std::cout << "proj%(id_proj)s_postevent: " << cudaGetErrorString(proj%(id_proj)s_postevent) << std::endl;
        }
    #endif
    }
"""
}

conn_templates = {
    # connectivity representation
    'conn_header': "const %(idx_type)s post_size, const %(idx_type)s pre_size, const char* mask",
    'conn_call': "pop%(id_post)s->size, pop%(id_pre)s->size, proj%(id_proj)s->device_mask()",
    'conn_kernel': "post_size, pre_size, mask",

    # launch config
    'launch_config': launch_config,

    # accessors
    'attribute_decl': attribute_decl,
    'attribute_cpp_init': attribute_cpp_init,
    'attribute_cpp_size': attribute_cpp_size,
    'attribute_cpp_delete': attribute_cpp_delete,
    'host_to_device': attribute_host_to_device,
    'device_to_host': attribute_device_to_host,
    'delay': delay,
    'event_driven': event_driven,

    #operations
    'rate_psp': rate_psp_kernel,
    'spike_transmission': {
        'event_driven': spike_event_transmission,
        'continuous': None
    },
    'synapse_update': {
        'global': global_synapse_update,
        'semiglobal': semiglobal_synapse_update,
        'local': local_synapse_update,
        'call': synapse_update_call
    },
    'post_event': spike_postevent
}

conn_ids = {
    'local_index': "[j]",
    'semiglobal_index': '[rk_post]',
    'global_index': '[0]',
    'pre_index': '[rk_pre]',
    'post_index': '[rk_post]',
    'pre_prefix': 'pre_',
    'post_prefix': 'post_'
}



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

attribute_decl = {
    'local':
"""
    // Local %(attr_type)s %(name)s
    std::vector<std::vector< %(type)s >> %(name)s;
""",
    'semiglobal':
"""
    // Semiglobal %(attr_type)s %(name)s
    std::vector< %(type)s >  %(name)s ;
""",
    'global':
"""
    // Global %(attr_type)s %(name)s
    %(type)s  %(name)s ;
"""
}

attribute_cpp_init = {
    'local':
"""
        // Local %(attr_type)s %(name)s
        %(name)s = init_matrix_variable<%(type)s>(static_cast<%(type)s>(%(init)s));
""",
    'semiglobal':
"""
        // Semiglobal %(attr_type)s %(name)s
        %(name)s = init_vector_variable<%(type)s>(static_cast<%(type)s>(%(init)s));
""",
    'global':
"""
        // Global %(attr_type)s %(name)s
        %(name)s = %(init)s;
"""
}

attribute_cpp_size = {
    'local': """
        // Local %(attr_type)s %(name)s
        size_in_bytes += sizeof(std::vector<std::vector<%(ctype)s>>);
        size_in_bytes += sizeof(std::vector<%(ctype)s>) * %(name)s.capacity();
        for(auto it = %(name)s.cbegin(); it != %(name)s.cend(); it++)
            size_in_bytes += (it->capacity()) * sizeof(%(ctype)s);
""",
    'semiglobal': """
        // Semiglobal %(attr_type)s %(name)s
        size_in_bytes += sizeof(std::vector<%(ctype)s>);
        size_in_bytes += sizeof(%(ctype)s) * %(name)s.capacity();
""",
    'global': """
        // Global
        size_in_bytes += sizeof(%(ctype)s);
"""
}

attribute_cpp_delete = {
    'local': """
        // %(name)s
        for (auto it = %(name)s.begin(); it != %(name)s.end(); it++) {
            it->clear();
            it->shrink_to_fit();
        };
        %(name)s.clear();
        %(name)s.shrink_to_fit();
""",
    'semiglobal': """
        // %(name)s
        %(name)s.clear();
        %(name)s.shrink_to_fit();
""",
    'global': ""
}

delay = {
    # A single value for all synapses
    'uniform': {
        'declare': """
    // Uniform delay
    int delay;

    int get_delay() { return delay; }
    int get_dendrite_delay(int idx) { return delay; }
    void set_delay(int delay) { this->delay = delay; }
""",
        'init': """
    delay = delays[0][0];
"""
    },
    # An individual value for each synapse
    'nonuniform_rate_coded': None,
    # An individual value for each synapse and a
    # buffer for spike events
    'nonuniform_spiking': None
}

continuous_transmission = {
    'sum': """
%(pre_copy)s

%(idx_type)s off, i, is, ie, j;
for (auto map_it = offsets_.begin(); map_it != offsets_.end(); map_it++) {

    off = map_it->first;
    is = std::max<int>(0, -off);
    ie = std::min<int>(num_columns_, num_columns_-off);

    %(omp_simd)s
    for(i=is; i < ie; i++) {
        %(post_prefix)s_sum_%(target)s%(post_index)s += %(psp)s;
    }
}
"""
}

conn_templates = {
    # accessors
    'attribute_decl': attribute_decl,
    'attribute_cpp_init': attribute_cpp_init,
    'attribute_cpp_size': attribute_cpp_size,
    'attribute_cpp_delete': attribute_cpp_delete,
    'delay': delay,

    'rate_coded_sum': continuous_transmission,
    'vectorized_default_psp': {},
    'update_variables': ""
}

conn_ids = {
    'local_index': '[map_it->second][i]',
    'semiglobal_index': '[i]',
    'global_index': '',
    'post_index': '[i]',
    'pre_index': '[i+off]',
    'delay_u' : '[delay-1]', # uniform delay
    'delay_nu' : '[delay[j]-1]' # nonuniform delay
}



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

import numpy as np
from ANNarchy.core.Random import RandomDistribution
from ANNarchy.intern import Messages


from functools import wraps
import time


######################
# Time measurement
######################
def timeit(func):
    """
    Decorator to measure the execution time of a method.

    ```python
    @ann.timeit
    def run(net, T):
        net.simulate(T)
        return net.m.get()

    net = ann.Network()
    data = run(net, 1000)
    ```
    """
    @wraps(func)
    def timeit_wrapper(*args, **kwargs):
        start_time = time.perf_counter()
        result = func(*args, **kwargs)
        end_time = time.perf_counter()
        total_time = end_time - start_time
        print(f'Function {func.__name__}{args} {kwargs} took {total_time:.4f} seconds')
        return result
    return timeit_wrapper

######################
# Sparse matrices
######################

def sparse_random_matrix(pre: "Population", post: "Population", proba:float, weights:float|RandomDistribution) -> "scipy.sparse.lil_matrix":
    """
    Returns a sparse lil-matrix for use in `Projection.from_sparse()`.
    
    Creates a `scipy.sparse.lil_matrix` connectivity matrix that connects the `pre` and `post` populations with the probability `p` and the value `weight`, which can be either a constant or an ANNarchy random distribution object.

    ```python
    pop1 = net.create(100, neuron)
    pop2 = net.create(100, neuron)
    proj = net.connect(pre=pop1, post=pop2, target='exc')

    matrix = sparse_random_matrix(pre=pop1, post=pop2, p=0.1, w=ann.Uniform(0.0, 1.0))
    proj.from_sparse(matrix)
    ```

    :param pre: pre-synaptic population.
    :param post: post-synaptic population.
    :param proba: connection probability.
    :param weights: weight values (constant or random).
    """
    try:
        from scipy.sparse import lil_matrix
    except:
        Messages._warning("scipy is not installed, sparse matrices won't work")
        return None
    
    from random import sample
    W = lil_matrix((pre.size, post.size))
    
    for i in range(pre.size):
        k=np.random.binomial(post.size, proba,1)[0]
        tmp = sample(range(post.size),k)
        W.rows[i]=list(np.sort(tmp))
        
        if isinstance(weights, (int, float)):
            W.data[i] = [weights]*k
        
        elif isinstance(weights, RandomDistribution):
            W.data[i] = weights.get_list_values(k)
        
        else:
            raise ValueError("sparse_random_matrix expects either a float or RandomDistribution object.")

    return W

def sparse_delays_from_weights(weights: "scipy.sparse.lil_matrix", delays: float | RandomDistribution) -> "scipy.sparse.lil_matrix":
    """
    Returns a sparse delay matrix with the same connectivity as the sparse matrix `weight_matrix`.

    ```python
    matrix = sparse_random_matrix(pre=pop1, post=pop2, p=0.1, w=ann.Uniform(0.0, 1.0))
    delays = sparse_delays_from_weights(matrix, ann.Uniform(5.0, 10.0))
    proj.from_sparse(matrix, delays)
    ```

    :param weights: scipy sparse matrix to use for the connectivity.
    :param delays: delay value (constant or random).
    """
    try:
        from scipy.sparse import lil_matrix
    except:
        Messages._warning("scipy is not installed, sparse matrices won't work")
        return None

    delay_matrix = lil_matrix(weights.get_shape())

    (rows, cols) = weights.nonzero()

    for r, c in zip(rows, cols):
        if isinstance(delays, (int, float)):
            delay_matrix[r,c] = delays
        elif isinstance(delays, RandomDistribution):
            delay_matrix[r,c] = delays.get_value()
        else:
            raise ValueError("sparse_random_matrix expects either a float or RandomDistribution object.")

    return delay_matrix

################################
## Performance Measurement
################################

def compute_delivered_spikes(proj, pre_spike_events, post_spike_events):
    """
    This function counts the number of delivered spikes for a given Projection
    considering both pre- and post-synaptic events.

    :param proj: the Projection.
    :param pre_spike_events: the pre-synaptic spike events per time step (the result of a spike recording)
    :param post_spike_events: the post-synaptic spike events per time step (the result of a spike recording)
    """
    pre_throughput = compute_delivered_efferent_spikes(proj, pre_spike_events)
    post_throughput = compute_delivered_afferent_spikes(proj, post_spike_events)
    return pre_throughput + post_throughput

def compute_delivered_efferent_spikes(proj, spike_events):
    """
    This function counts the number of delivered spikes for a given Projection and
    spike sequence.

    :param proj: the Projection.
    :param spike_events: the spike events per time step (the result of a spike recording)
    """
    nb_efferent_synapses = proj._nb_efferent_synapses()
    delivered_events = 0

    for neur_rank, time_steps in spike_events.items():
        if neur_rank in proj.pre.ranks and neur_rank in nb_efferent_synapses.keys():
            delivered_events += nb_efferent_synapses[neur_rank] * len(time_steps)

    return delivered_events

def compute_delivered_afferent_spikes(proj, spike_events):
    """
    This function counts the number of delivered spikes for a given Projection and
    spike sequence.

    :param proj: the Projection.
    :param spike_events: the spike events per time step (the result of a spike recording)
    """
    delivered_events = 0

    for neur_rank, time_steps in spike_events.items():
        if neur_rank in proj.post.ranks:
            delivered_events += proj.dendrite(neur_rank).size * len(time_steps)

    return delivered_events

def compute_delivered_spikes_per_second(proj, spike_events, time_in_seconds, scale_factor=(10**6)):
    """
    This function implements a throughput metric for spiking neural networks.

    :param proj: the Projection
    :param spike_events: the spike events per time step (the result of a spike recording)
    :param time_in_seconds: computation time used for the operation
    :param scale_factor: usually the throughput value gets quite large, therefore its useful to rescale the value. By default, we re-scale to millions of events per second.

    **Note**:

    In the present form the function should be used only if a single projection is recorded. If multiple Projections contribute to the
    measured time (*time_in_seconds*) you need to compute the value individually for each Projection using the *compute_delivered_spikes*
    method and the times for instance retrieved from a run using --profile.
    """
    num_events = compute_delivered_spikes(proj, spike_events)

    return (num_events / time_in_seconds) / scale_factor



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

attribute_decl = {
    'local':
"""
    // Local %(attr_type)s %(name)s
    std::vector<%(type)s> %(name)s;
""",
    'semiglobal':
"""
    // Semiglobal %(attr_type)s %(name)s
""",
    'global':
"""
    // Global %(attr_type)s %(name)s
"""
}

attribute_cpp_init = {
    'local':
"""
        // Local %(attr_type)s %(name)s
        w = init_matrix_variable<%(type)s>(%(init)s);
""",
    'semiglobal':
"""
        // Semiglobal %(attr_type)s %(name)s
""",
    'global':
"""
        // Global %(attr_type)s %(name)s
"""
}

attribute_cpp_size = {
    'local': """
        // Local %(attr_type)s %(name)s
        size_in_bytes += sizeof(std::vector<%(ctype)s>);
        size_in_bytes += sizeof(%(ctype)s) * %(name)s.capacity();
""",
    'semiglobal': """
        // Semiglobal %(attr_type)s %(name)s
""",
    'global': """
        // Global %(attr_type)s %(name)s
"""
}

attribute_cpp_delete = {
    'local': """
        // %(name)s
""",
    'semiglobal': """
        // %(name)s
""",
    'global': ""
}

#############################################
##  Synaptic delay
#############################################
delay = {
    # A single value for all synapses
    'uniform': None,
    # An individual value for each synapse
    'nonuniform_rate_coded': None,
    # An individual value for each synapse and a
    # buffer for spike events
    'nonuniform_spiking': None
}

###############################################################
# Rate-coded continuous transmission (default implementation)
###############################################################
continuous_transmission = {
    'sum' : """
    %(idx_type)s row_ptr_size = block_row_size();
    %(idx_type)s row_max = this->get_num_rows();
    %(idx_type)s tile_size = get_tile_size();
    %(idx_type)s tile_size2 = tile_size*tile_size;
    auto block_ptr = block_row_pointer();
    auto block_col_idx = block_column_index();

    %(float_prec)s* __restrict__ target_ptr = %(post_prefix)s_sum_%(target)s.data();
    %(float_prec)s* __restrict__ pre_r = %(pre_prefix)sr.data();

    %(omp_code)s %(omp_clause)s %(omp_schedule)s
    for (%(idx_type)s blk_row = 0; blk_row < row_ptr_size; blk_row++) {
        %(idx_type)s row_nr_off = blk_row * tile_size;
        %(float_prec)s* loc_psp = target_ptr + row_nr_off;

        for (%(idx_type)s blk_col_idx = block_ptr[blk_row]; blk_col_idx < block_ptr[blk_row+1]; blk_col_idx++) {
            %(idx_type)s bcol_idx = block_col_idx[blk_col_idx];     // which column in row

            %(float_prec)s* __restrict__ values = w.data() + blk_col_idx * tile_size2;       // find the correct dense tile
            %(float_prec)s* __restrict__ loc_pr = pre_r + bcol_idx * tile_size;              // select the correct part in pre vector

            // process the dense matrix row by row
            %(idx_type)s row_off = 0;
            %(idx_type)s r = 0;
            for (; r < tile_size; r++, row_off+=tile_size) {
                if (row_nr_off+r >= row_max)
                    continue;

                %(float_prec)s sum = 0.0;

                // process all columns in the row
                for (%(idx_type)s c = 0; c < tile_size; c++) {
                    sum += values[row_off+c] * loc_pr[c];
                }
                loc_psp[r] += sum;
            }
        }
    }
"""
}

# The inner dense SpMV is partially unrolled (2x2 kernel)
continuous_transmission_unroll_2x2 = {
    'sum' : """
        %(idx_type)s row_ptr_size = block_row_size();
        %(idx_type)s row_max = this->get_num_rows();
        %(idx_type)s tile_size = get_tile_size();
        %(idx_type)s tile_size2 = tile_size*tile_size;
        auto block_ptr = block_row_pointer();
        auto block_col_idx = block_column_index();

        %(float_prec)s* __restrict__ target_ptr = %(post_prefix)s_sum_%(target)s.data();
        %(float_prec)s* __restrict__ pre_r = %(pre_prefix)sr.data();

        #pragma omp for
        for (%(idx_type)s blk_row = 0; blk_row < row_ptr_size; blk_row++) {
            %(idx_type)s row_nr_off = blk_row * tile_size;
            %(float_prec)s* loc_psp = target_ptr + row_nr_off;
            %(float_prec)s sum1=0.0;
            %(float_prec)s sum2=0.0;

            /* Unrolled inner 2x2 SpMV kernel */
            for (%(idx_type)s blk_col_idx = block_ptr[blk_row]; blk_col_idx < block_ptr[blk_row+1]; blk_col_idx++) {
                %(idx_type)s bcol_idx = block_col_idx[blk_col_idx];     // which column in row

                %(float_prec)s* __restrict__ values = w.data() + blk_col_idx * tile_size2;       // find the correct dense tile
                %(float_prec)s* __restrict__ loc_pr = pre_r + bcol_idx * tile_size;              // select the correct part in pre vector

                sum1 += values[0] * loc_pr[0];
                sum1 += values[1] * loc_pr[1];

                sum2 += values[2] * loc_pr[0];
                sum2 += values[3] * loc_pr[1];
            }

            // prevent wrong write access
            if (row_nr_off < row_max)
                loc_psp[0] += sum1;
            if (row_nr_off+1 < row_max)
                loc_psp[1] += sum2;

        }
"""
}

# The inner dense SpMV is partially unrolled (3x3 kernel)
continuous_transmission_unroll_3x3 = {
    'sum' : """
        %(idx_type)s row_ptr_size = block_row_size();
        %(idx_type)s row_max = this->get_num_rows();
        %(idx_type)s tile_size = get_tile_size();
        %(idx_type)s tile_size2 = tile_size*tile_size;
        auto block_ptr = block_row_pointer();
        auto block_col_idx = block_column_index();

        %(float_prec)s* __restrict__ target_ptr = %(post_prefix)s_sum_%(target)s.data();
        %(float_prec)s* __restrict__ pre_r = %(pre_prefix)sr.data();

        #pragma omp for
        for (%(idx_type)s blk_row = 0; blk_row < row_ptr_size; blk_row++) {
            %(idx_type)s row_nr_off = blk_row * tile_size;
            %(float_prec)s* loc_psp = target_ptr + row_nr_off;

            for (%(idx_type)s blk_col_idx = block_ptr[blk_row]; blk_col_idx < block_ptr[blk_row+1]; blk_col_idx++) {
                %(idx_type)s bcol_idx = block_col_idx[blk_col_idx];     // which column in row

                %(float_prec)s* __restrict__ values = w.data() + blk_col_idx * tile_size2;       // find the correct dense tile
                %(float_prec)s* __restrict__ loc_pr = pre_r + bcol_idx * tile_size;              // select the correct part in pre vector

                %(float_prec)s sum1_1, sum1_2, sum1_3, sum2_1, sum2_2, sum2_3, sum3_1, sum3_2, sum3_3;
                sum1_1 = values[0] * loc_pr[0];
                sum1_2 = values[1] * loc_pr[1];
                sum1_3 = values[2] * loc_pr[2];

                sum2_1 = values[4] * loc_pr[0];
                sum2_2 = values[5] * loc_pr[1];
                sum2_3 = values[6] * loc_pr[2];

                sum3_1 = values[7] * loc_pr[0];
                sum3_2 = values[8] * loc_pr[1];
                sum3_3 = values[9] * loc_pr[2];

                if (row_nr_off < row_max)
                    loc_psp[0] += sum1_1 + sum1_2 + sum1_3;
                if (row_nr_off+1 < row_max)
                    loc_psp[1] += sum2_1 + sum2_2 + sum2_3;
                if (row_nr_off+2 < row_max)
                    loc_psp[2] += sum3_1 + sum3_2 + sum3_3;
            }
        }
"""
}

###############################################################
# Rate-coded synaptic plasticity
###############################################################
update_variables = {
    'local': ""
}

conn_templates = {
    # accessors
    'delay': delay,
    'attribute_decl': attribute_decl,
    'attribute_cpp_init': attribute_cpp_init,
    'attribute_cpp_size': attribute_cpp_size,
    'attribute_cpp_delete': attribute_cpp_delete,
    
    'rate_coded_sum': continuous_transmission,
    # optimized kernels sorted by block size and SIMD
    'unrolled_default_psp': {
        2: {
            'none': {
                'multi_w': continuous_transmission_unroll_2x2
            },
            #'sse', 'avx'
        },
        3: {
            'none': {
                'multi_w': continuous_transmission_unroll_3x3,
            }
            #'sse','avx'
        }
    },
    'update_variables': update_variables
}

conn_ids = {
    'local_index': '',
    'pre_index': '',
    'post_index': '',
}


"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.intern.SpecificProjection import SpecificProjection
from ANNarchy.intern.ConfigManagement import ConfigManager

from ANNarchy.intern import Messages

class AccProjection(SpecificProjection):
    """
    Accumulates the values of a given variable.
    """
    def __init__(self, pre, post, target, variable, name=None, normalize_input=0, scale_factor=1.0, copied=False, net_id=0):
        
        # Instantiate the projection
        SpecificProjection.__init__(self, pre, post, target, None, name, copied, net_id)
        
        self._variable = variable
        self._scale_factor = scale_factor
        self._normalize_input = normalize_input

        # Check population type of the receiving population
        if not self.post.neuron_type.type == 'rate':
            Messages._error('The post-synaptic population of an AccProjection must be rate-coded.')

        # Prevent automatic split of matrices
        self._no_split_matrix = True

    def _copy(self, pre, post, net_id=None):
        "Returns a copy of the population when creating networks. Internal use only."
        return AccProjection(
            pre=pre, post=post, target=self.target, 
            variable=self._variable, 
            name=self.name, normalize_input=self._normalize_input, 
            scale_factor=self._scale_factor, 
            copied=True, 
            net_id=self.net_id if net_id is None else net_id)

    def _generate_st(self):
        """
        """
        # Sanity Check
        found = False
        for var in self.pre.neuron_type.description['variables']:
            if var['name'] == self._variable:
                found = True
                break

        if not found:
            Messages._warning("Variable might be invalid ...")

        single_ids = {'id_pre': self.pre.id,'var': self._variable}

        if self._normalize_input == 0:
            # Generate Code Template
            self._specific_template['psp_prefix'] = ""
            self._specific_template['psp_code'] = """
        for(int post_idx = 0; post_idx < post_rank.size(); post_idx++) {
        %(float_prec)s lsum = 0.0;

            for(auto it = pre_rank[post_idx].begin(); it != pre_rank[post_idx].end(); it++) {
                lsum += pop%(id_pre)s->%(var)s[*it];
            }

            pop%(id_post)s->_sum_%(target)s[post_rank[post_idx]] += %(scale_factor)s * lsum/pre_rank[post_idx].size();
        }
""" % {
    'id_post': self.post.id,
    'id_pre': self.pre.id,
    'var': self._variable,
    'target': self.target,
    'scale_factor': self._scale_factor,
    'float_prec': ConfigManager().get('precision', self.net_id)
}

        else:
            # Generate Code Template
            self._specific_template['declare_additional'] = """
    std::vector<std::vector<%(float_prec)s>> baseline;
    std::vector<%(float_prec)s> baseline_mean;
    std::vector<%(float_prec)s> baseline_std;
    long time_for_init_baseline;
    int init_baseline_period;
    void start(int baseline_period) {
        init_baseline_period=baseline_period;
        time_for_init_baseline = t + baseline_period;
    #ifdef _DEBUG
        std::cout << "ProjStruct%(id_proj)s: set new baseline period from step " << t << " to step " << time_for_init_baseline << std::endl;
    #endif
    }
""" % {'id_proj': self.id, 'float_prec': ConfigManager().get('precision', self.net_id)}
            self._specific_template['export_additional'] = """
        void start(int)
"""

            self._specific_template['init_additional'] = """
        time_for_init_baseline = -1;
        init_baseline_period=1;
        baseline = std::vector<std::vector<%(float_prec)s>>(post_rank.size(), std::vector<%(float_prec)s>() );
        baseline_mean = std::vector<%(float_prec)s>(post_rank.size(), 0);
        baseline_std = std::vector<%(float_prec)s>(post_rank.size(), 1);
""" % {'float_prec': ConfigManager().get('precision', self.net_id)}
            
            self._specific_template['clear_additional'] = """
        for(auto it = baseline.begin(); it != baseline.end(); it++) {
            it->clear();
            it->shrink_to_fit();
        }
        baseline.clear();
        baseline.shrink_to_fit();

        baseline_mean.clear();
        baseline_mean.shrink_to_fit();

        baseline_std.clear();
        baseline_std.shrink_to_fit();
"""
            self._specific_template['psp_prefix'] = ""
            self._specific_template['psp_code'] = """
        bool compute_baseline = (t < time_for_init_baseline) ? true : false;
        bool compute_average = (t == time_for_init_baseline) ? true : false;

        for(int post_idx = 0; post_idx < post_rank.size(); post_idx++) {
            %(float_prec)s lsum = 0.0;

            // accumulate the input variable
            auto it = pre_rank[post_idx].begin();
            int j = 0;
            for(; it != pre_rank[post_idx].end(); it++, j++) {
                lsum += pop%(id_pre)s->%(var)s[*it];
            }

            // we want to use the average across incoming connections
            // ( i. e. the recorded population )
            lsum /= static_cast<%(float_prec)s>(pre_rank[post_idx].size());

            // if the init time is over compute the mean/standard
            // deviation across time
            if (compute_average) {
                %(float_prec)s sum = std::accumulate(std::begin(baseline[post_idx]), std::end(baseline[post_idx]), static_cast<%(float_prec)s>(0.0));
                baseline_mean[post_idx] = sum / static_cast<%(float_prec)s>(baseline[post_idx].size());

                %(float_prec)s accum = 0.0;
                std::for_each (baseline[post_idx].begin(), baseline[post_idx].end(), [&](const %(float_prec)s value) {
                    accum += (value - baseline_mean[post_idx]) * (value - baseline_mean[post_idx]);
                });
                baseline_std[post_idx] = sqrt(accum / static_cast<%(float_prec)s>(baseline[post_idx].size()-1));
            }

            // until init time is reached we store the rescaled sum,
            // otherwise we use the baseline mean and standard deviation to rescale
            if (compute_baseline) {
                // enqueue the value to baseline vector
                baseline[post_idx].push_back(lsum);

                // don't store the result
                pop%(id_post)s->_sum_%(target)s[post_rank[post_idx]] += 0.0;
            } else {
                // apply relative deviation normalization
                lsum = ((lsum - baseline_mean[post_idx]) / (std::abs(baseline_mean[post_idx]) + 0.0000001));

                // store the result
                pop%(id_post)s->_sum_%(target)s[post_rank[post_idx]] += %(scale_factor)s * lsum;
            }
        }
""" % {
    'id_post': self.post.id,
    'id_pre': self.pre.id,
    'var': self._variable,
    'target': self.target,
    'scale_factor': self._scale_factor,
    'float_prec': ConfigManager().get('precision', self.net_id)
}
            
            self._specific_template['wrapper'] = f"""
    // AccProjection ProjStruct{self.id}
    nanobind::class_<ProjStruct{self.id}>(m, "proj{self.id}_wrapper")
        // Constructor
        .def(nanobind::init<>())

        // Flags
        .def_rw("_transmission", &ProjStruct{self.id}::_transmission)
        .def_rw("_axon_transmission", &ProjStruct{self.id}::_axon_transmission)
        .def_rw("_update", &ProjStruct{self.id}::_update)
        .def_rw("_plasticity", &ProjStruct{self.id}::_plasticity)

        // Connectivity
        .def("init_from_lil", &ProjStruct{self.id}::init_from_lil)
        .def("post_rank", &ProjStruct{self.id}::get_post_rank)
        .def("dendrite_size", &ProjStruct{self.id}::dendrite_size)
        .def("pre_ranks", &ProjStruct{self.id}::get_pre_ranks)
        .def("pre_rank", &ProjStruct{self.id}::get_dendrite_pre_rank)

        // Start
        .def("start", &ProjStruct{self.id}::start)

        // Attributes
        .def("get_global_attribute_{ConfigManager().get('precision', self.net_id)}", &ProjStruct{self.id}::get_global_attribute_{ConfigManager().get('precision', self.net_id)})
        .def("set_global_attribute_{ConfigManager().get('precision', self.net_id)}", &ProjStruct{self.id}::set_global_attribute_{ConfigManager().get('precision', self.net_id)})


        // Other methods
        .def("clear", &ProjStruct{self.id}::clear);
        """

    def _generate_omp(self):
        """
        """
        # Sanity Check
        found = False
        for var in self.pre.neuron_type.description['variables']:
            if var['name'] == self._variable:
                found = True
                break

        if not found:
            Messages._warning("Variable might be invalid ...")

        single_ids = {'id_pre': self.pre.id,'var': self._variable}

        if self._normalize_input == 0:
            # Generate Code Template
            self._specific_template['psp_prefix'] = ""
            self._specific_template['psp_code'] = """
        #pragma omp for
        for(int post_idx = 0; post_idx < post_rank.size(); post_idx++) {
            %(float_prec)s lsum = 0.0;

            for(auto it = pre_rank[post_idx].begin(); it != pre_rank[post_idx].end(); it++) {
                lsum += pop%(id_pre)s->%(var)s[*it];
            }

            pop%(id_post)s->_sum_%(target)s[post_rank[post_idx]] += %(scale_factor)s * lsum/pre_rank[post_idx].size();
        }
""" % {
    'id_post': self.post.id,
    'id_pre': self.pre.id,
    'var': self._variable,
    'target': self.target,
    'scale_factor': self._scale_factor,
    'float_prec': ConfigManager().get('precision', self.net_id)
}

        else:
            # Generate Code Template
            self._specific_template['declare_additional'] = """
    std::vector<std::vector<%(float_prec)s>> baseline;
    std::vector<%(float_prec)s> baseline_mean;
    std::vector<%(float_prec)s> baseline_std;
    long time_for_init_baseline;
    int init_baseline_period;
    void start(int baseline_period) {
        init_baseline_period=baseline_period;
        time_for_init_baseline = t + baseline_period;
    #ifdef _DEBUG
        std::cout << "ProjStruct%(id_proj)s: set new baseline period from step " << t << " to step " << time_for_init_baseline << std::endl;
    #endif
    }
""" % {'id_proj': self.id, 'float_prec': ConfigManager().get('precision', self.net_id)}
            self._specific_template['export_additional'] = """
        void start(int)
"""

            self._specific_template['init_additional'] = """
        time_for_init_baseline = -1;
        init_baseline_period=1;
        baseline = std::vector<std::vector<%(float_prec)s>>(post_rank.size(), std::vector<%(float_prec)s>() );
        baseline_mean = std::vector<%(float_prec)s>(post_rank.size(), 0);
        baseline_std = std::vector<%(float_prec)s>(post_rank.size(), 1);
""" % {'float_prec': ConfigManager().get('precision', self.net_id)}

            self._specific_template['psp_prefix'] = ""
            self._specific_template['psp_code'] = """
        bool compute_baseline = (t < time_for_init_baseline) ? true : false;
        bool compute_average = (t == time_for_init_baseline) ? true : false;

        #pragma omp for
        for(int post_idx = 0; post_idx < post_rank.size(); post_idx++) {
            %(float_prec)s lsum = 0.0;

            // accumulate the input variable
            auto it = pre_rank[post_idx].begin();
            int j = 0;
            for(; it != pre_rank[post_idx].end(); it++, j++) {
                lsum += pop%(id_pre)s->%(var)s[*it];
            }

            // we want to use the average across incoming connections
            // ( i. e. the recorded population )
            lsum /= static_cast<%(float_prec)s>(pre_rank[post_idx].size());

            // if the init time is over compute the mean/standard
            // deviation across time
            if (compute_average) {
                %(float_prec)s sum = std::accumulate(std::begin(baseline[post_idx]), std::end(baseline[post_idx]), static_cast<%(float_prec)s>(0.0));
                baseline_mean[post_idx] = sum / static_cast<%(float_prec)s>(baseline[post_idx].size());

                %(float_prec)s accum = 0.0;
                std::for_each (baseline[post_idx].begin(), baseline[post_idx].end(), [&](const %(float_prec)s value) {
                    accum += (value - baseline_mean[post_idx]) * (value - baseline_mean[post_idx]);
                });
                baseline_std[post_idx] = sqrt(accum / static_cast<%(float_prec)s>(baseline[post_idx].size()-1));
            }

            // until init time is reached we store the rescaled sum,
            // otherwise we use the baseline mean and standard deviation to rescale
            if (compute_baseline) {
                // enqueue the value to baseline vector
                baseline[post_idx].push_back(lsum);

                // don't store the result
                pop%(id_post)s._sum_%(target)s[post_rank[post_idx]] += 0.0;
            } else {
                // apply relative deviation normalization
                lsum = ((lsum - baseline_mean[post_idx]) / (std::abs(baseline_mean[post_idx]) + 0.0000001));

                // store the result
                pop%(id_post)s->_sum_%(target)s[post_rank[post_idx]] += %(scale_factor)s * lsum;
            }
        }
""" % {
    'id_post': self.post.id,
    'id_pre': self.pre.id,
    'var': self._variable,
    'target': self.target,
    'scale_factor': self._scale_factor,
    'float_prec': ConfigManager().get('precision', self.net_id)
}

    def _generate_cuda(self):
        raise NotImplementedError("The AccProjection (part of the BOLD monitor) is not available for CUDA devices yet.")



from .SingleThreadGenerator import SingleThreadGenerator as SingleThreadProjectionGenerator
from .OpenMPGenerator import OpenMPGenerator as OpenMPProjectionGenerator
from .CUDAGenerator import CUDAGenerator as CUDAProjectionGenerator


"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

import ANNarchy.core.Global as Global

from ANNarchy.intern import Messages

try:
    from tensorboardX import SummaryWriter
except Exception as e:
    print(e)
    Messages._error("tensorboard extension: please install tensorboardX (pip install tensorboardX).")

import os
import socket
from datetime import datetime
import numpy as np


class Logger(object):
    """
    Logger class to use tensorboard to visualize ANNarchy simulations. Requires the `tensorboardX` package (pip install tensorboardX). 

    The Logger class is a thin wrapper around tensorboardX.SummaryWriter, which you could also use directly. The doc is available at <https://tensorboardx.readthedocs.io/>. Tensorboard can read any logging data, as long as they are saved in the right format (tfevents), so it is not limited to tensorflow. TensorboardX has been developed to allow the use of tensorboard with pytorch.

    The extension has to be imported explictly:

    ```python
    from ANNarchy.extensions.tensorboard import Logger
    ```

    The ``Logger`` class has to be closed properly at the end of the script, so it is advised to use a context:

    ```python
    with Logger() as logger:
        logger.add_scalar("Accuracy", acc, trial)
    ```

    You can also make sure to close it:

    ```python
    logger = Logger()
    logger.add_scalar("Accuracy", acc, trial)
    logger.close()
    ```

    By default, the logs will be written in a subfolder of ``./runs/`` (which will be created in the current directory). 
    The subfolder is a combination of the current datetime and of the hostname, e.g. ``./runs/Apr22_12-11-22_machine``. 
    You can control these two elements by passing arguments to ``Logger()``:

    ```python
    with Logger(logdir="/tmp/annarchy", experiment="trial1"): # logs in /tmp/annarchy/trial1
    ```

    The ``add_*`` methods allow you to log various structures, such as scalars, images, histograms, figures, etc.

    A tag should be given to each plot. In the example above, the figure with the accuracy will be labelled "Accuracy" in tensorboard. 
    You can also group plots together with tags such as "Global performance/Accuracy", "Global performance/Error rate", "Neural activity/Population 1", etc.

    After (or while) logging data within your simulation, run `tensorboard` in the terminal by specifying the log directory:

    ```bash
    tensorboard --logdir runs
    ```

    TensorboardX enqueues the data in memory before writing to disk. You can force flushing with:

    ```python
    logger.flush()
    ```

    :param logdir: path (absolute or relative) to the logging directory. Subfolders will be created for each individual run. The default is "runs/"
    :param experiment: name of the subfolder for the current run. By default, it is a combination of the current time and the hostname (e.g. Apr22_12-11-22_machine). If you reuse an experiment name, the data will be appended.
    """
    
    def __init__(self, logdir:str="runs/", experiment:str=None):

        self.logdir = logdir
        self.experiment = experiment

        # Create the logdir if it does not exist
        if not os.path.exists(self.logdir):
            os.makedirs(self.logdir)

        if not experiment:
            current_time = datetime.now().strftime('%b%d_%H-%M-%S')
            self.currentlogdir=os.path.join(
                self.logdir, current_time + '_' + socket.gethostname())
        else:
            self.currentlogdir = self.logdir + "/" + self.experiment

        print("Logging in", self.currentlogdir)
    
        self._create_summary_writer()

    def _create_summary_writer(self):

         self._summary = SummaryWriter(self.currentlogdir, comment="", purge_step=None, max_queue=10, flush_secs=10, filename_suffix='', write_to_disk=True)

    # Logging methods
        
    def add_scalar(self, tag:str, value:float, step:int=None):
        """
        Logs a single scalar value, e.g. a success rate at various stages of learning.

        Example:

        ```python
        with Logger() as logger:
            for trial in range(100):
                simulate(1000.0)
                accuracy = ...
                logger.add_scalar("Accuracy", accuracy, trial)
        ```

        :param tag: name of the figure in tensorboard.
        :param value: value.
        :param step: time index.
        """
        
        self._summary.add_scalar(tag=tag, scalar_value=value, global_step=step, walltime=None)
        
    def add_scalars(self, tag:str, value:dict, step:int=None):
        """
        Logs multiple scalar values to be displayed in the same figure, e.g. several metrics or neural activities.

        Example:

        ```python
        with Logger() as logger:
            for trial in range(100):
                simulate(1000.0)
                act1 = pop.r[0]
                act2 = pop.r[1]
                logger.add_scalars(
                    "Accuracy", 
                    {'First neuron': act1, 'Second neuron': act2}, 
                    trial)
        ```

        :param tag: name of the figure in tensorboard.
        :param value: dictionary of values.
        :param step: time index.
        """
        
        self._summary.add_scalars(main_tag=tag, tag_scalar_dict=value, global_step=step, walltime=None)
        
    def add_image(self, tag:str, img: np.ndarray, step:int=None, equalize:bool=False):
        """
        Logs an image.
        
        The image must be a numpy array of size (height, width) for monochrome images or (height, width, 3) for colored images. The values should either be integers between 0 and 255 or floats between 0 and 1. The parameter ``equalize`` forces the values to be between 0 and 1 by equalizing using the min/max values.

        Example::

        ```python
        with Logger() as logger:
            for trial in range(100):
                simulate(1000.0)
                img = pop.r.reshape((10, 10))
                logger.add_image("Population / Firing rate", img, trial, equalize=True)
        ```

        :param tag: name of the figure in tensorboard.
        :param img: array for the image.
        :param step: time index.
        :param equalize: rescales the pixels between 0 and 1 using the min and max values of the array.
        """
        if img.ndim ==2:
            if equalize:  
                img = img.astype(np.float)              
                img = (img - img.min())/(img.max() - img.min())

            self._summary.add_image(tag=tag, img_tensor=img, global_step=step, walltime=None, dataformats='HW')
        
        elif img.ndim == 3:
            if not img.shape[2] == 3:
                Messages._error("Logger.add_image: color images must be of shape (H, W, 3).")
            
            if equalize:   
                img = np.array(img).astype(np.float)         
                img = (img - img.min())/(img.max() - img.min())

            self._summary.add_image(tag=tag, img_tensor=img, global_step=step, walltime=None, dataformats='HWC')

        else:
            Messages._error("Logger.add_image: images must be of shape (H, W) or (H, W, 3).")
        
    def add_images(self, tag:str, img:np.array, step:int=None, equalize:bool=False, equalize_per_image:bool=False):
        """
        Logs a set of images (e.g. receptive fields).
       
        The numpy array must be of size (number, height, width) for monochrome images or (number, height, width, 3) for colored images. The values should either be integers between 0 and 255 or floats between 0 and 1. The parameter ``equalize`` forces the values to be between 0 and 1 by equalizing using the min/max values.

        Example:

        ```python
        with Logger() as logger:
            for trial in range(100):
                simulate(1000.0)
                weights= proj.w.reshape(100, 10, 10) # 100 post neurons, 10*10 pre neurons
                logger.add_images("Projection/Receptive fields", weights, trial, equalize=True)
        ```

        :param tag: name of the figure in tensorboard.
        :param img: array for the images.
        :param step: time index.
        :param equalize: rescales the pixels between 0 and 1 using the min and max values of the array.
        :param equalize_per_image: whether the rescaling should be using the global min/max values of the array, or per image. Has no effect if equalize of False.
 
        """
        if img.ndim == 3:
            img = np.expand_dims(img, axis=3)
        
        if equalize:   
            img = np.array(img).astype(np.float) 
            if not equalize_per_image:        
                img = (img - img.min())/(img.max() - img.min())
            else:
                for i in range(img.shape[0]):
                    img[i,...] = (img[i,...] - img[i,...].min())/(img[i,...].max() - img[i,...].min())
        
        self._summary.add_images(tag=tag, img_tensor=img, global_step=step, walltime=None, dataformats='NHWC')
        
    def add_parameters(self, params:dict, metrics:dict):
        """
        Logs parameters of a simulation.

        This should be run only once per simulation, generally at the end. 
        This allows to compare different runs of the same network using 
        different parameter values and study how they influence the global output metrics, 
        such as accuracy, error rate, reaction speed, etc.

        Example:

        ```python
        with Logger() as logger:
            # ...
            logger.add_parameters({'learning_rate': lr, 'tau': tau}, {'accuracy': accuracy})
        ```

        :param params: dictionary of parameters.
        :param metrics: dictionary of metrics.
        """
        
        self._summary.add_hparams(params, metrics)

    def add_histogram(self, tag:str, hist: list | np.ndarray, step:int=None):
        """
        Logs an histogram.

        Example:

        ```python
        with Logger() as logger:
            for trial in range(100):
                simulate(1000.0)
                weights= proj.w.flatten()
                logger.add_histogram("Weight distribution", weights, trial)
        ```


        :param tag: name of the figure in tensorboard.
        :param hist: a list or 1D numpy array of values.
        :param step: time index.
        """

        self._summary.add_histogram(tag, hist, step)

    def add_figure(self, tag:str, figure: list | np.ndarray, step:int=None, close:bool=True):
        """
        Logs a Matplotlib figure.

        Example:

        ```python
        with Logger() as logger:
            for trial in range(100):
                simulate(1000.0)
                fig = plt.figure()
                plt.plot(pop.r)
                logger.add_figure("Activity", fig, trial)
        ```

        :param tag: name of the image in tensorboard.
        :param figure: a list or 1D numpy array of values.
        :param step: time index.
        :param close: whether the logger will close the figure when done (default: True).
        """

        import matplotlib.pyplot as plt
        import matplotlib.backends.backend_agg as plt_backend_agg
        canvas = plt_backend_agg.FigureCanvasAgg(figure)
        canvas.draw()
        data = np.frombuffer(canvas.buffer_rgba(), dtype=np.uint8)
        w, h = figure.canvas.get_width_height()
        image_hwc = data.reshape([h, w, 4])[:, :, 0:3]
        image_chw = np.moveaxis(image_hwc, source=2, destination=0)
        if close:
            plt.close(figure)
        self._summary.add_image(tag, image_chw, step)
    
    # Resource management
    def flush(self):
        "Forces the logged data to be flushed to disk."
        self._summary.flush()
        
    def close(self):
        "Closes the logger."
        self._summary.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

attribute_decl = {
    'local':
"""
    // Local %(attr_type)s %(name)s
    std::vector< %(type)s > %(name)s;
""",
    'semiglobal':
"""
    // Semiglobal %(attr_type)s %(name)s
    std::vector< %(type)s >  %(name)s ;
""",
    'global':
"""
    // Global %(attr_type)s %(name)s
    %(type)s  %(name)s ;
"""
}

attribute_cpp_init = {
    'local':
"""
        // Local %(attr_type)s %(name)s
        %(name)s = init_matrix_variable<%(type)s>(static_cast<%(type)s>(%(init)s));
""",
    'semiglobal':
"""
        // Semiglobal %(attr_type)s %(name)s
        %(name)s = init_vector_variable<%(type)s>(static_cast<%(type)s>(%(init)s));
""",
    'global':
"""
        // Global %(attr_type)s %(name)s
        %(name)s = %(init)s;
"""
}

attribute_cpp_size = {
    'local': """
        // Local %(attr_type)s %(name)s
        size_in_bytes += sizeof(std::vector<%(ctype)s>);
        size_in_bytes += sizeof(%(ctype)s) * %(name)s.capacity();       
""",
    'semiglobal': """
        // Semiglobal %(attr_type)s %(name)s
        size_in_bytes += sizeof(std::vector<%(ctype)s>);
        size_in_bytes += sizeof(%(ctype)s) * %(name)s.capacity();
""",
    'global': """
        // Global
        size_in_bytes += sizeof(%(ctype)s);
"""
}

attribute_cpp_delete = {
    'local': """
        // %(name)s
        %(name)s.clear();
        %(name)s.shrink_to_fit();
""",
    'semiglobal': """
        // %(name)s
        %(name)s.clear();
        %(name)s.shrink_to_fit();
""",
    'global': ""
}

delay = {
    'uniform': {
        'declare': """
    // Uniform delay
    int delay ;""",

        'pyx_struct':
"""
        # Uniform delay
        int delay""",
        'init': """
    delay = delays[0][0];
""",
        'pyx_wrapper_init':
"""
        proj%(id_proj)s.delay = syn.uniform_delay""",
        'pyx_wrapper_accessor':
"""
    # Access to non-uniform delay
    def get_delay(self):
        return proj%(id_proj)s.delay
    def get_dendrite_delay(self, idx):
        return proj%(id_proj)s.delay
    def set_delay(self, value):
        proj%(id_proj)s.delay = value
"""
    }
}

###############################################################
# Rate-coded continuous transmission
###############################################################
ellr_summation_operation = {
    'sum' : """
%(pre_copy)s

%(float_prec)s* __restrict__ target = %(post_prefix)s_sum_%(target)s.data();
%(idx_type)s nb_post = static_cast<%(idx_type)s>(post_ranks_.size());

%(omp_code)s %(omp_clause)s %(omp_schedule)s
for(%(idx_type)s i = 0; i < nb_post; i++) {
    rk_post = post_ranks_[i]; // Get postsynaptic rank

    sum = 0.0;
    for(%(size_type)s j = i*maxnzr_; j < i*maxnzr_+rl_[i]; j++) {
        rk_pre = col_idx_[j];
        sum += %(psp)s ;
    }
    
    target%(post_index)s += sum;
}""",
    'max': "",
    'min': "",
    'mean': ""
}

###############################################################
# Rate-coded synaptic plasticity
###############################################################
update_variables = {
    'local': """
// Check periodicity
if(_transmission && _update && %(post_prefix)s_active && ( (t - _update_offset)%%_update_period == 0L) ){
    // Global variables
    %(global)s

    %(idx_type)s nb_post = static_cast<%(idx_type)s>(post_ranks_.size());
    // Local variables
    #pragma omp for
    for(%(idx_type)s i = 0; i < nb_post; i++){
        rk_post = post_ranks_[i]; // Get postsynaptic rank
        // Semi-global variables
        %(semiglobal)s
        // Local variables
        for(%(size_type)s j = i*maxnzr_; j < i*maxnzr_+rl_[i]; j++){
            rk_pre = col_idx_[j]; // Get presynaptic rank
    %(local)s
        }
    }
}
"""
}

conn_templates = {
    # accessors
    'attribute_decl': attribute_decl,
    'attribute_cpp_init': attribute_cpp_init,
    'attribute_cpp_size': attribute_cpp_size,
    'attribute_cpp_delete': attribute_cpp_delete,
    'delay': delay,
    
    'rate_coded_sum': ellr_summation_operation,
    'update_variables': update_variables
}

conn_ids = {
    'local_index': '[j]',
    'semiglobal_index': '[i]',
    'global_index': '',
    'post_index': '[rk_post]',
    'pre_index': '[rk_pre]',
    'delay_u' : '[delay-1]' # uniform delay
}



"""
Contains functions for load/save of parameters, connectivtiy and complete networks.

:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.core import Global
from ANNarchy.core.Constant import Constant
from ANNarchy.intern.NetworkManager import NetworkManager
from ANNarchy.intern import Messages

import re
import os
import pickle
import numpy as np


def load_parameters(filename:str, global_only:bool=True, verbose:bool=False, net_id:int=0) -> dict:
    """
    Loads the global parameters of a network (flag ``population`` for neurons, ``projection`` for synapses) from a JSON file.

    It returns a dictionary of additional parameters not related to populations or projections (keyword ``network`` in the JSON file).

    It is advised to generate the JSON file first with `save_parameters()` and later edit it manually.

    A strong restriction is that population/projection names cannot change between saving and loading.
    By default, they take names such as ``pop0`` or ``proj2``, we advise setting explicitly a name in their constructor for readability.

    If you add a parameter name to the JSON file but it does not exist in te neuron/synapse, it will be silently skipped.
    Enable ``verbose=True`` to see which parameters are effectively changed.

    If you set ``global_only`` to True, you will be able to set values for non-global parameters (e.g. synapse-specific), but a single value will be loaded for all.
    The JSON file cannot contain arrays.

    If you want to save/load the value of variables after a simulation, please refer to ``save()`` or ``load()``.

    :param filename: path to the JSON file.
    :param global_only: True if only global parameters (flags ``population`` and ``projection``) should be loaded, the other values are ignored. (default: True)
    :param verbose: True if the old and new values of the parameters should be printed (default: False).

    """
    import json
    try:
        with open(filename, 'r') as rfile:
            desc = json.load(rfile)
    except IOError as error:
        print(error)
        Messages._error("load_parameters(): the json file does not exist")   

    if verbose:
        Messages._print('Loading parameters from file', filename)
        Messages._print('-'*40)

    # Populations
    try:
        populations = desc['populations']
    except:
        populations = {}
        if verbose:
            Messages._print('load_parameters(): no population parameters.')
    for name, parameters in populations.items():
        # Get the population
        for pop in NetworkManager().get_network(net_id=net_id).get_populations():
            if pop.name == name:
                population = pop
                break
        else:
            Messages._warning('The population', name, 'defined in the file', filename, 'does not exist in the current network.')

        if verbose:
            Messages._print('Population', name)

        # Set the parameters
        for name, val in parameters.items():
            # Check that the variable indeed exists
            if not name in population.parameters:
                Messages._print('  ', name, 'is not a global parameter of', population.name, ', skipping.')
                continue
            if global_only and not name in population.neuron_type.description['global']:
                Messages._print('  ', name, 'is not a global parameter of', population.name, ', skipping.')
                continue

            if verbose:
                Messages._print('  ', name, ':', population.get(name), '->', val)

            population.set({name: val})

    # Projections
    try:
        projections = desc['projections']
    except:
        projections = {}
        if verbose:
            Messages._print('load_parameters(): no projection parameters.')
    for name, parameters in projections.items():
        # Get the projection
        for proj in NetworkManager().get_network(net_id=net_id).get_projections():
            if proj.name == name:
                projection = proj
                break
        else:
            Messages._warning('The projection', name, 'defined in the file', filename, 'does not exist in the current network.')

        if verbose:
            Messages._print('Projection', name)

        # Set the parameters
        for name, val in parameters.items():
            # Check that the variable indeed exists
            if not name in projection.parameters:
                Messages._print('  ', name, 'is not a global parameter of', population.name, ', skipping.')
                continue
            if global_only and not name in projection.synapse_type.description['global']:
                Messages._print('  ', name, 'is not a global parameter of', population.name, ', skipping.')
                continue

            if verbose:
                Messages._print('  ', name, ':', projection.get(name), '->', val)

            projection.set({name: float(val)})

    # Constants
    try:
        constants = desc['constants']
    except:
        constants = {}
        if verbose:
            Messages._print('load_parameters(): no constants.')
    for name, value in constants.items():
        # Get the constant
        for constant in NetworkManager().get_network(net_id=net_id).get_constants():
            if constant.name == name:
                constant.set(value)
                break
        else:
            Messages._warning('The projection', name, 'defined in the file', filename, 'does not exist in the current network.')

    # Global user-defined parameters
    try:
        network_parameters = {}
        for name, val in desc['network'].items():
            network_parameters[name] = float(val)
    except:
        network_parameters = {}

    return network_parameters

def save_parameters(filename:str, net_id=0):
    """
    Saves the global parameters of a network (flag ``population`` for neurons, ``projection`` for synapses) to a JSON file.

    :param filename: path to the JSON file.
    """
    import json

    # Get the network description
    network = NetworkManager().get_network(net_id=net_id)

    # Dictionary of parameters
    description = {
        'populations' : {},
        'projections' : {},
        'network' : {},
        'constants' : {},
    }

    # Constants
    for constant in network.get_constants():
        description['constants'][constant.name] = constant.value

    # Populations
    for pop in network.get_populations():

        # Get the neuron description
        neuron = pop.neuron_type

        pop_description = {}

        for param in neuron.description['global']:
            pop_description[param] = pop.init[param]

        description['populations'][pop.name] = pop_description

    # Projections
    for proj in network.get_projections():

        # Get the synapse description
        synapse = proj.synapse_type

        proj_description = {}

        for param in synapse.description['global']:
            proj_description[param] = proj.init[param]

        description['projections'][proj.name] = proj_description

    # Save the description in a json file
    try:
        with open(filename, 'w') as wfile:
            json.dump(description, wfile, indent=4)
    except IOError as error:
        Messages._error("save_parameters(): cannot write the json file. Make sure the subfolders already exist.")


# Backwards compatibility with XML
def load_parameter(in_file):
    Messages._warning('load_parameter() is deprecated. Use load_parameters() and JSON files instead.')
    return _load_parameters_from_xml(in_file)

def _load_parameters_from_xml(in_file):
    """
    Load parameter set from xml file.

    If the location of the xml file differs from the base directory, you need to provide relative or absolute path.

    :param in_file: either single or collection of strings.
    """
    try:
        from lxml import etree
    except:
        Messages._print('lxml is not installed. Unable to load in xml format.')
        return
    par = {}
    damaged_pars = []   # for printout

    files = []
    if isinstance(in_file,str):
        files.append(in_file)
    else:
        files = in_file

    for file in files:
        try:
            doc = etree.parse(file)

        except IOError:
            Messages._print('Error: file \'', file, '\' not found.')
            continue

        matches = doc.findall('parameter')

        for parameter in matches:
            childs = parameter.getchildren()

            #TODO: allways correct ???
            if len(childs) != 2:
                Messages._print('Error: to much tags in parameter')

            name=None
            value=None
            for child in childs:

                if child.tag == 'name':
                    name = child.text
                elif child.tag == 'value':
                    value = child.text

                    if value is None:
                        Messages._print('Error: no value defined for',name)
                        damaged_pars.append(name)
                        value = 0
                    else:
                        try:
                            value = int(value)
                        except ValueError:
                            try:
                                value = float(value)
                            except ValueError:
                                value = value

                else:
                    Messages._print('Error: unexpected xml-tag', child.tag)

            if name is None:
                Messages._print('Error: no name in parameter set.')
            elif value is None:
                Messages._print('Error: no value in parameter set.')
                damaged_pars.append(name)
            elif name in par.keys():
                Messages._print("Error: parameter",name,"already exists.")
                damaged_pars.append(name)
            else:
                par[name] = value

    return par

def _save_data(filename, data):
    """
    Internal routine to save data in a file.

    """
    # Check if the repertory exist
    (path, fname) = os.path.split(filename)

    if not path == '':
        if not os.path.isdir(path):
            Messages._print('Creating folder', path)
            os.mkdir(path)

    extension = os.path.splitext(fname)[1]

    if extension == '.mat':
        Messages._print("Saving network in Matlab format...")
        try:
            import scipy.io as sio
            sio.savemat(filename, data)
        except Exception as e:
            Messages._error('Error while saving in Matlab format.')
            Messages._print(e)
            return

    elif extension == '.gz':
        Messages._print("Saving network in gunzipped binary format...")
        try:
            import gzip
        except:
            Messages._error('gzip is not installed.')
            return
        with gzip.open(filename, mode = 'wb') as w_file:
            try:
                pickle.dump(data, w_file, protocol=pickle.HIGHEST_PROTOCOL)
            except Exception as e:
                Messages._print('Error while saving in gzipped binary format.')
                Messages._print(e)
                return

    elif extension == '.npz':
        Messages._print("Saving network in Numpy format...")
        np.savez_compressed(filename, allow_pickle=True, **data )

    else:
        Messages._print("Saving network in text format...")
        # save in Pythons pickle format
        with open(filename, mode = 'wb') as w_file:
            try:
                pickle.dump(data, w_file, protocol=pickle.HIGHEST_PROTOCOL)
            except Exception as e:
                Messages._print('Error while saving in text format.')
                Messages._print(e)
                return
        return

def save(filename:str, populations:bool=True, projections:bool=True, net_id=0) -> None :
    """
    Save the current network state (parameters and variables) to a file.

    * If the extension is '.npz', the data will be saved and compressed using `np.savez_compressed` (recommended).

    * If the extension is '.mat', the data will be saved as a Matlab 7.2 file. Scipy must be installed.

    * If the extension ends with '.gz', the data will be pickled into a binary file and compressed using gzip.

    * Otherwise, the data will be pickled into a simple binary text file using cPickle.

    **Warning:** The '.mat' data will not be loadable by ANNarchy, it is only for external analysis purpose.

    Example:

    ```python
    ann.save('results/init.npz')

    ann.save('results/init.data')

    ann.save('results/init.txt.gz')

    ann.save('1000_trials.mat')
    ```

    :param filename: filename, may contain relative or absolute path.
    :param populations: if True, population data will be saved (by default True)
    :param projections: if True, projection data will be saved (by default True)
    """
    data = _net_description(populations, projections, net_id)
    _save_data(filename, data)

def _load_data(filename, pickle_encoding):
    """
    Internally loads data contained in a given file.

    Returns a dictionary with the connectivity and synaptic variables if the file ``filename`` is available otherwise None is returned.

    :param filename: path to the file.
    :param pickle_encoding: if set to None the default is used, e.g. Python2 files ("latin1") or Python3 files ("ASCII")
    """
    (_, fname) = os.path.split(filename)
    extension = os.path.splitext(fname)[1]

    if extension == '.mat':
        Messages._error('Unable to load Matlab format.')
        return None

    elif extension == '.gz':
        try:
            import gzip
        except:
            Messages._error('gzip is not installed.')
            return None
        try:
            with gzip.open(filename, mode = 'rb') as r_file:
                if pickle_encoding is None:
                    desc = pickle.load(r_file)
                else:
                    desc = pickle.load(r_file, encoding=pickle_encoding)
            return desc
        except Exception as e:
            Messages._print('Unable to read the file ' + filename)
            Messages._print(e)
            return None

    elif extension == '.npz':
        try:
            if pickle_encoding is None:
                data = np.load(filename, allow_pickle=True)
            else:
                data = np.load(filename, allow_pickle=True, encoding=pickle_encoding)
            desc = {}
            for attribute in data.files:
                # We need to distinguish two cases: 1) full network save
                # or 2) single pop/proj. The first case leads to a dictionary
                # of several objects. The latter to a dictionary containing all
                # values.
                if data[attribute].dtype == np.dtype('O'):
                    # attribute is a collection of multiple objects
                    desc[attribute] = data[attribute].item(0)
                else:
                    # attribute is a scalar/array
                    desc[attribute] = data[attribute]

            return desc
        except Exception as e:
            Messages._print('Unable to read the file ' + filename)
            Messages._print(e)
            return None

    else:
        try:
            with open(filename, mode = 'rb') as r_file:
                if pickle_encoding is None:
                    desc = pickle.load(r_file)
                else:
                    desc = pickle.load(r_file, encoding=pickle_encoding)
            return desc
        except Exception as e:
            Messages._print('Unable to read the file ' + filename)
            Messages._print(e)
            return None

def _load_connectivity_data(filename, pickle_encoding):
    """
    Internally loads data contained in a given file.

    Returns a dictionary with the connectivity and synaptic variables if the file ``filename`` is available otherwise None is returned.

    :param filename: path to the file.
    :param pickle_encoding: Pickle encoding.
    """
    (_, fname) = os.path.split(filename)
    extension = os.path.splitext(fname)[1]

    if extension == '.mat':
        Messages._error('Unable to load Matlab format.')
        return None

    elif extension == '.gz':
        try:
            import gzip
        except:
            Messages._error('gzip is not installed.')
            return None
        try:
            with gzip.open(filename, mode = 'rb') as r_file:
                if pickle_encoding is None:
                    desc = pickle.load(r_file)
                else:
                    desc = pickle.load(r_file, encoding=pickle_encoding)
            return desc
        except Exception as e:
            Messages._print('Unable to read the file ' + filename)
            Messages._print(e)
            return None

    elif extension == '.npz':
        try:
            if pickle_encoding is None:
                data = np.load(filename, allow_pickle=True)
            else:
                data = np.load(filename, allow_pickle=True, encoding=pickle_encoding)
            desc = {}
            for attribute in data.files:
                # We need to distinguish two cases: 1) full network save
                # or 2) single pop/proj. The first case leads to a dictionary
                # of several objects. The latter to a dictionary containing all
                # values.
                desc[attribute] = data[attribute]

            return desc
        except Exception as e:
            Messages._print('Unable to read the file ' + filename)
            Messages._print(e)
            return None

    else:
        try:
            with open(filename, mode = 'rb') as r_file:
                if pickle_encoding is None:
                    desc = pickle.load(r_file)
                else:
                    desc = pickle.load(r_file, encoding=pickle_encoding)
            return desc
        except Exception as e:
            Messages._print('Unable to read the file ' + filename)
            Messages._print(e)
            return None

def load(filename:str, populations:bool=True, projections:bool=True, pickle_encoding:str=None, net_id=0):
    """
    Loads a saved state of the network.

    **Warning:** Matlab data can not be loaded.

    Example:

    ```python
    ann.load('results/network.npz')
    ```

    :param filename: the filename with relative or absolute path.
    :param populations: if True, population data will be loaded (by default True)
    :param projections: if True, projection data will be loaded (by default True)
    :param pickle_encoding: optional parameter provided to the pickle.load() method. If set to None the default is used.
    """

    desc = _load_data(filename, pickle_encoding)
    if desc is None:
        return

    if 'time_step' in desc.keys():
        Global.set_current_step(desc['time_step'], net_id)

    if populations:
        # Over all populations
        for pop in NetworkManager().get_network(net_id=net_id).get_populations():
            # check if the population is contained in save file
            if pop.name in desc.keys():
                pop._load_pop_data(desc[pop.name])

    if projections:
        for proj in NetworkManager().get_network(net_id=net_id).get_projections():
            if proj.name in desc.keys():
                proj._load_proj_data(desc[proj.name])


def _net_description(populations, projections, net_id=0):
    """
    Returns a dictionary containing the requested network data.

    :param populations: if True, the population data will be saved.
    :param projections: if True, the projection data will be saved.
    """
    network_desc = {}
    network_desc['time_step'] = Global.get_current_step(net_id)
    network_desc['net_id'] = net_id

    pop_names = []
    proj_names = []

    if populations:
        for pop in NetworkManager().get_network(net_id=net_id).get_populations():
            network_desc[pop.name] = pop._data()
            pop_names.append(pop.name)

    if projections:
        for proj in NetworkManager().get_network(net_id=net_id).get_projections():
            # Some specific projections are note saveable
            if not proj._saveable:
                continue
            network_desc[proj.name] = proj._data()
            proj_names.append(proj.name)

    network_desc['obj_names'] = {
        'populations': pop_names,
        'projections': proj_names,
    }

    return network_desc


class MonitorList(list):
    """
    A helper object to gather monitors for easy saving as a h5py file.

    Example:

    ```python
    monitors = MonitorList([
        Monitor(pop1, ['r', 'sum(exc)']),
        Monitor(pop2, 'r')
    ])

    monitors.save_all("monitors.hdf5")
    ```
    """
    def save_all(self, filename: str, keep: bool = False, reshape: bool = False, flat_keys: bool = False) -> None:
        """
        Saves all the recorded variables as a h5py file of numpy arrays.

        The `spike` variable of a population will be returned as a dictionary of lists,
        where the spike times (in steps) for each recorded neurons are saved.

        :param filename: name of the save file.
        :param keep: defines if the content in memory for each variable should be kept (default: False).
        :param reshape: transforms the second axis of the array to match the population's geometry (default: False).
        :param flat_keys: flatten keys in the saved file. This might be useful if you track just a few parameters (default: False).

        """
        # Define the separator for the keys
        if not filename.endswith(".hdf5"):
            Messages._error('MonitorList: File type must be .hdf5-File.')

        try:
            import h5py
        except Exception as e:
            Messages._error('MonitorList: h5py is required to save the monitors, please install it.')


        if flat_keys:
            sep = "_"
        else:
            sep = "/"

        with h5py.File(filename, 'w') as data:
            for mon in self:
                for var in mon.variables:
                    name = var
                    # Sums of inputs for rate-coded populations
                    if var.startswith('sum('):
                        target = re.findall(r"\(([\w]+)\)", var)[0]
                        name = '_sum_' + target

                    # Retrieve the data
                    data[f"/{mon.object.name}{sep}{var}"] = mon._return_variable(name, keep, reshape)

                    # Update stopping time
                    mon._update_stopping_time(var, keep)



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

import numpy as np

from ANNarchy.intern import Messages
from ANNarchy.intern.ConfigManagement import get_global_config

distributions_arguments = {
    'Uniform' : 2,
    'DiscreteUniform': 2,
    'Normal' : 2,
    'LogNormal': 2,
    'Exponential': 1,
    'Gamma': 2,
    'Binomial' : 2
}

distributions_equivalents = {
    'Uniform' : 'std::uniform_real_distribution< %(float_prec)s >',
    'DiscreteUniform': 'std::uniform_int_distribution<int>',
    'Normal' : 'std::normal_distribution< %(float_prec)s >',
    'LogNormal': 'std::lognormal_distribution< %(float_prec)s >',
    'Exponential': 'std::exponential_distribution< %(float_prec)s >',
    'Gamma': 'std::gamma_distribution< %(float_prec)s >',
    'Binomial': 'std::binomial_distribution<int>'
}

# List of available distributions
available_distributions = list(distributions_arguments.keys())

class RandomDistribution :
    """
    BaseClass for random distributions.
    """

    def get_values(self, shape):
        """
        Returns a np.ndarray with the given shape
        """
        Messages._error('instantiated base class RandomDistribution is not allowed.')
        return np.array([0.0])

    def get_list_values(self, size):
        """
        Returns a list of the given size.
        """
        return list(self.get_values(size))

    def get_value(self):
        """
        Returns a single float value.
        """
        return self.get_values(1)[0]

    def keywords(self):
        return available_distributions

    def latex(self):
        return '?'
    
    def get_cpp_args(self):
        raise NotImplementedError

class Uniform(RandomDistribution):
    """
    Uniform distribution between ``min`` and ``max``.

    The returned values are floats in the range [min, max].

    :param min: minimum value.
    :param max: maximum value.
    :param rng: (optional) random number generator. If left None, `np.random.default_rng()` is used.
    """
    def __init__(self, min: float, max:float, rng:np.random.Generator=None):

        if rng is None:
            self.rng = np.random.default_rng()
        else:
            self.rng = rng

        if min >= max:
            Messages._error(f"Uniform: the minimum value (min = {min}) should be smaller than the maximum value (max = {max}).")

        self.min = min
        self.max = max

    def get_values(self, shape:tuple) -> np.ndarray:
        """
        Returns a Numpy array with the given shape.

        :param shape: shape of the array.
        """
        return self.rng.uniform(self.min, self.max, shape)

    def latex(self):
        return "$\\mathcal{U}$(" + str(self.min) + ', ' + str(self.max) + ')'

    def get_cpp_args(self):
        return self.min, self.max

class DiscreteUniform(RandomDistribution):
    """
    Discrete uniform distribution between ``min`` and ``max``.

    The returned values are integers in the range [min, max].

    :param min: minimum value.
    :param max: maximum value.
    :param rng: (optional) random number generator. If left None, `np.random.default_rng()` is used.
    """
    def __init__(self, min: int, max:int, rng:np.random.Generator=None):

        if rng is None:
            self.rng = np.random.default_rng()
        else:
            self.rng = rng

        if min >= max:
            Messages._error(f"Uniform: the minimum value (min = {min}) should be smaller than the maximum value (max = {max}).")

        self.min = min
        self.max = max

    def get_values(self, shape:tuple) -> np.ndarray:
        """
        Returns a Numpy array with the given shape.

        :param shape: shape of the array.
        """
        # randint draws from half-open interval [min, max)
        return self.rng.integers(self.min, self.max+1, shape)

    def latex(self):
        return "$\\mathcal{U}$(" + str(self.min) + ', ' + str(self.max) + ')'


class Normal(RandomDistribution):
    """
    Normal distribution.

    :param mu: Mean of the distribution.
    :param sigma: Standard deviation of the distribution.
    :param min: Minimum value (default: unlimited).
    :param max: Maximum value (default: unlimited).
    :param rng: (optional) random number generator. If left None, `np.random.default_rng()` is used.
    """
    def __init__(self, mu:float, sigma:float, min:float=None, max:float=None, rng:np.random.Generator=None):

        if rng is None:
            self.rng = np.random.default_rng()
        else:
            self.rng = rng

        if sigma < 0.0:
            Messages._error("Normal: the standard deviation sigma should be positive.")

        self.mu = mu
        self.sigma = sigma
        self.min = min
        self.max = max

    def get_values(self, shape:tuple) -> np.ndarray:
        """
        Returns a Numpy array with the given shape.

        :param shape: shape of the array.
        """
        data = self.rng.normal(self.mu, self.sigma, shape)

        if self.min != None:
            data[data<self.min] = self.min
        if self.max != None:
            data[data>self.max] = self.max
        return data

    def latex(self):
        return "$\\mathcal{N}$(" + str(self.mu) + ', ' + str(self.sigma) + ')'

    def get_cpp_args(self):
        return self.mu, self.sigma

class LogNormal(RandomDistribution):
    """
    Log-normal distribution.

    :param mu: Mean of the distribution.
    :param sigma: Standard deviation of the distribution.
    :param min: Minimum value (default: unlimited).
    :param max: Maximum value (default: unlimited).
    :param rng: (optional) random number generator. If left None, `np.random.default_rng()` is used.
    """
    def __init__(self, mu:float, sigma:float, min:float=None, max:float=None, rng:np.random.Generator=None):

        if rng is None:
            self.rng = np.random.default_rng()
        else:
            self.rng = rng

        if sigma < 0.0:
            Messages._error("LogNormal: the standard deviation sigma should be positive.")

        self.mu = mu
        self.sigma = sigma
        self.min = min
        self.max = max

    def get_values(self, shape:tuple) -> np.ndarray:
        """
        Returns a Numpy array with the given shape.

        :param shape: shape of the array.
        """
        data = self.rng.lognormal(self.mu, self.sigma, shape)

        if self.min != None:
            data[data<self.min] = self.min
        if self.max != None:
            data[data>self.max] = self.max
        return data

    def latex(self):
        return "$\\ln\\mathcal{N}$(" + str(self.mu) + ', ' + str(self.sigma) + ')'

    def get_cpp_args(self):
        return self.mu, self.sigma

class Exponential(RandomDistribution):
    """
    Exponential distribution, according to the density function:

    $$P(x | \\lambda) = \\lambda e^{(-\\lambda x )}$$

    **Note:** ``Lambda`` is capitalized, otherwise it would be a reserved Python keyword.

    :param Lambda: rate parameter.
    :param min: minimum value (default: unlimited).
    :param max: maximum value (default: unlimited).
    :param rng: (optional) random number generator. If left None, `np.random.default_rng()` is used.
    """
    def __init__(self, Lambda:float, min:float=None, max:float=None, rng:np.random.Generator=None):

        if rng is None:
            self.rng = np.random.default_rng()
        else:
            self.rng = rng

        if Lambda < 0.0:
            Messages._error("Exponential: the rate parameter Lambda should be positive.")

        self.Lambda = Lambda
        self.min = min
        self.max = max

    def get_values(self, shape:tuple) -> np.ndarray:
        """
        Returns a Numpy array with the given shape.

        :param shape: shape of the array.
        """
        data = self.rng.exponential(self.Lambda, shape)

        if self.min != None:
            data[data<self.min] = self.min
        if self.max != None:
            data[data>self.max] = self.max
        return data

    def latex(self):
        return "$\\exp$(" + str(self.Lambda) + ')'

class Gamma(RandomDistribution):
    """
    Gamma distribution.

    :param alpha: Shape of the gamma distribution.
    :param beta: Scale of the gamma distribution.
    :param min: Minimum value returned (default: unlimited).
    :param max: Maximum value returned (default: unlimited).
    :param rng: (optional) random number generator. If left None, `np.random.default_rng()` is used.
    """
    def __init__(self, alpha:float, beta:float=1.0, min:float=None, max:float=None, rng:np.random.Generator=None):

        if rng is None:
            self.rng = np.random.default_rng()
        else:
            self.rng = rng

        self.alpha = alpha
        self.beta = beta
        self.min = min
        self.max = max

    def get_values(self, shape:tuple) -> np.ndarray:
        """
        Returns a Numpy array with the given shape.

        :param shape: shape of the array.
        """
        data = self.rng.gamma(self.alpha, self.beta, shape)

        if self.min != None:
            data[data<self.min] = self.min
        if self.max != None:
            data[data>self.max] = self.max
        return data

    def latex(self):
        return "$\\Gamma$(" + str(self.alpha) + ', ' + str(self.beta) + ')'
        
class Binomial(RandomDistribution):
    """
    Binomial distribution.
    
    Parameters: n trials and p probability of success where n an integer >= 0 and p is in the interval [0,1].

    The returned values are the number of successes over the n trials.

    :param n: Number of trials.
    :param p: Probability of success.
    :param rng: (optional) random number generator. If left None, `np.random.default_rng()` is used.
    """

    def __init__(self, n:int, p:float, rng:np.random.Generator=None):

        if rng is None:
            self.rng = np.random.default_rng()
        else:
            self.rng = rng

        self.n = n
        self.p = p

    def get_values(self, shape:tuple) -> np.ndarray:
        """
        Returns a Numpy array with the given shape.

        :param shape: shape of the array.
        """
        return self.rng.binomial(self.n, self.p, size=shape)

    def latex(self):
        return "$\\mathcal{B}$(" + str(self.n) + ", " + str(self.p) + ")"

    def get_cpp_args(self):
        return self.n, self.p



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from dataclasses import dataclass
from ANNarchy.core.Random import RandomDistribution

# Parameters

@dataclass
class Parameter:
    """
    Dataclass to represent a parameter in a Neuron or Synapse definition.

    ```python
    neuron = ann.Neuron(
        parameters = dict(

            # Global parameter
            tau = 10.0 # or ann.Parameter(value=10.0, locality='global')

            # Local parameter
            baseline = ann.Parameter(value=ann.Uniform(-1., 1.)),

            # Boolean global parameter
            activated = ann.Parameter(value=True, locality='global', type=bool),
        )
    )
    ```

    In a neuron or synapse model, parameters are global and use the float type if the `Parameter` class is not used. 
    
    If you need a local parameter (one value per neuron or synapse), the `Parameter` class allows to specify it. Note that you can also define global parameters by passing `locality='global'`.

    Semi-global synaptic parameters (one value per post-synaptic neuron) can be defined using `locality='semiglobalglobal'`.

    If the parameter is an int or a bool, pass it to the `type` attribute.

    :param value: Initial value of the parameter. It can be defined as a RandomDistribution, which will be sampled with the correct shape when the population/projection is created, or a float/int/bool, depending on `type`.
    :param locality: Locality of the parameter. Must be in ['global', 'semiglobal', 'local'].
    :param type: Data type of the parameter. Must be in [float, int, bool] (or ['float', 'int', 'bool']).
    """

    value: float | int | bool | RandomDistribution
    locality: str = 'local'
    type: str = 'float'


# Variables
@dataclass
class Variable:
    """
    Dataclass to represent a variable in a Neuron or Synapse definition.

    ```python
    neuron = ann.Neuron(
        equations = [
            ann.Variable('C * dv/dt = - gL * (v - E_L) +  gL * delta_T * exp((v-v_T)/delta_T) + I - w', init=-70.0),

            ann.Variable('tau_w * dw/dt = a * (v - E_L) - w', min=0.0),
        ]
    )
    ```

    Variables are local by default. Set `locality` to `global` or `semiglobal` otherwise.

    :param equation: string representing the equation.
    :param init: initial value of the variable. It can be defined as a RandomDistribution, which will be sampled with the correct shape when the population/projection is created, or a float/int/bool, depending on `type`.
    :param min: minimum value that the variable can take. 
    :param max: maximum value that the variable can take. 
    :param method: numerical method to be used when the equation is an ODE. Must be in ['explicit', 'implicit', 'semiimplicit', 'exponential','midpoint', 'rk4', 'event-driven']
    :param locality: Locality of the parameter. Must be in ['global', 'semiglobal', 'local'].
    :param type: Data type of the parameter. Must be in [float, int, bool] (or ['float', 'int', 'bool']).
    """

    equation: str
    init: float | int | bool | RandomDistribution = None
    min: float = None
    max: float = None
    method: str = None
    locality: str = 'local'
    type: str = 'float'

    def _to_string(self, object_type:str) -> str:
        "Returns a one-liner string with the flags. object_type is either 'neuron' or 'synapse'. to decide between population and projection"

        representation = self.equation

        # Any flag?
        has_flag = False
        flags = []

        if self.init is not None:
            if not has_flag: has_flag = True ; representation += " : "
            if isinstance(self.init, (RandomDistribution,)):
                init = repr(self.init)
            else: 
                init = self.init
            flags.append(f"init={init}") 

        if self.min is not None:
            if not has_flag: has_flag = True ; representation += " : "
            flags.append(f"min={self.min}") 

        if self.max is not None:
            if not has_flag: has_flag = True ; representation += " : "
            flags.append(f"max={self.max}") 

        if self.method is not None:
            if not has_flag: has_flag = True ; representation += " : "
            flags.append(f"{self.method}") 

        if self.type != 'float':
            if not has_flag: has_flag = True ; representation += " : "
            flags.append(f"{self.type}") 

        if self.locality != 'local':
            if not has_flag: has_flag = True ; representation += " : "
            if object_type == 'neuron':
                val = 'population' # 'global' and 'population' would work
            elif object_type == 'synapse':
                if self.locality in ['global', 'projection']: # both are OK
                    val = 'projection'
                else: val = 'postsynaptic'
            flags.append(f"{val}") 


        return representation + ', '.join(flags)
    

# Structural plasticity
@dataclass
class Creating:
    """
    Dataclass to represent a creation condition for structural plasticity.

    When the condition is true, a synapse is created with the specified probability, using the weight $w$ and delay $d$.

    ```python
    CreatingSynapse = ann.Synapse(

        parameters = dict(eta = 0.1, T = 1.0),
    
        equations = ann.Variable("dw/dt = eta * pre.r * post.r"),
        
        creating = ann.Creating("pre.r * post.r > T", proba = 0.1, w = 0.01),
    )
    ```

    :param equation: string representing the equation.
    :param proba: probability of creation of the synapse.
    :param w: weight when the synapse is created.
    :param d: delay when the synapse is created.
    """
    equation: str
    proba: float = 1.0
    w: float = 0.0
    d: float = None 

@dataclass
class Pruning:
    """
    Dataclass to represent a pruning condition for structural plasticity.

    When the condition is true, a synapse is pruned with the specified probability.

    ```python
    PruningSynapse = ann.Synapse(
        parameters = dict(T = ann.Parameter(10000, 'global', int),
        equations = ann.Variable('''
            age = if pre.r * post.r > 0.0 : 
                    0
                else :
                    age + 1 : init = 0, int
        ''', init=0, type=int)
        pruning = ann.Pruning("age > T", proba = 0.5),
    )
    ```

    :param equation: string representing the equation.
    :param proba: probability of pruning of the synapse.

    """
    equation: str
    proba: float = 1.0


"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from typing import Union
import numpy as np
import copy

from ANNarchy.intern.NetworkManager import NetworkManager
from ANNarchy.intern import Messages

default_config = dict(
    # Simulation Control
    dt = 1.0,
    seed = None,
    method = 'explicit',
    structural_plasticity = False,
    # Parallel processing
    num_threads = 1,
    visible_cores = [],
    paradigm = 'openmp',
    # Datatype-related
    precision = "double",
    only_int_idx_type = True,
    # Logging
    verbose = False,
    suppress_warnings = False,
    show_time = False,
    trace_calls = None,
    # Performance-related
    disable_parallel_rng = True,
    use_seed_seq = True,
    use_cpp_connectors = False,
    disable_split_matrix = True,
    disable_SIMD_SpMV = True,
    disable_SIMD_Eq = False,
    # SpM formats
    sparse_matrix_format = "default",
    sparse_matrix_storage_order = "post_to_pre",
    # Profiling
    profiling = False,
    profile_out = None,
    # Other
    debug = False,
    disable_shared_library_time_offset = True
)

class ConfigManager:
    """
    Manages the global configuration flags used in the ANNarchy framework. Users can manipulate this
    flags via two globally available functions:

        *setup()*
        *_optimization_flags()*

    Implementation Note:

        The class is implemented as singleton to ensure unique existance in the user space.
        One should not access the _config member directly but using the get/set methods.
    """
    _instance = None
    
    def __new__(self, *args, **kwds):
        """
        Only the first call will create a new instance of this class.
        """
        if self._instance is None:
            self._instance = super().__new__(self, *args, **kwds)
            self._config = { # dictionary because network ids can change
                0: copy.deepcopy(default_config)
            }

            # This flags can not be configured through setup()
            self._performance_related_config_keys = [
                'use_cpp_connectors', 'disable_split_matrix', 'disable_SIMD_SpMV', 'disable_SIMD_Eq'
            ]

        return self._instance
    
    def register_network(self, net_id:int=0):
        """
        Registers a network by copying the "magic" configuration of id 0.
        """
        self._config[net_id] = copy.deepcopy(self._config[0])
    
    def get_config(self, net_id:int=0):
        """
        Returns the config for the given network.
        """
        return copy.deepcopy(self._config[net_id])
    
    def set_config(self, net_id:int, config):
        """
        Set the config for the given network.
        """
        if not net_id in self._config.keys():
            self.register_network(net_id)
        self._config[net_id].update(config)

    
    def get(self, key: str, net_id:int=0) -> str | float | bool:
        """
        Returns the configuration for entry `key` in the network `net_id`. 
        
        If the key does not exist, a terminating exception is raised.
        """
        if key in self.keys():
            return self._config[net_id][key]
        else:
            raise Messages.ANNarchyException(key, "does not belong to global configuration keys.", exit=True)
        

    def set(self, key: str, value: str | float | bool, net_id:int=0):
        """
        Updates the configuration for entry *key* with a new *value*. 
        If the key does not exist a terminating exception is raised.
        """
        if key in self.keys():
            self._config[net_id][key] = value
        else:
            raise KeyError
        
    def keys(self) -> list:
        "Returns the list of keys that can be set with setup."
        return list(self._config[0].keys())


#############################################
# Globally available functions (internal)
#############################################
def setup(**keyValueArgs):
    """
    The setup function is used to configure ANNarchy simulations. 
    
    It takes various optional arguments. The most useful ones are:

    * `dt`: simulation step size in milliseconds (default: 1.0).
    * `paradigm`: parallel framework for code generation. Accepted values: "openmp" or "cuda" (default: "openmp").
    * `method`: default method to numerize ODEs. Default is the explicit forward Euler method ('explicit').
    * `precision`: default floating precision for variables in ANNarchy. Accepted values: "float" or "double" (default: "double")
    * `structural_plasticity`: allows synapses to be dynamically added/removed during the simulation (default: False).
    * `seed`: the seed (integer) to be used in the random number generators (default = None is equivalent to time(NULL)).
    * `num_threads`: number of treads used by openMP (overrides the environment variable ``OMP_NUM_THREADS`` when set, default = None).

    Flags related to the optimization of the simulation kernels are:

    * `sparse_matrix_format`: the default matrix format for projections in ANNarchy (by default: List-In-List for CPUs and Compressed Sparse Row). Note that this affects only the C++ data structures.
    * `sparse_matrix_storage_order`: encodes whether the row in a connectivity matrix encodes pre-synaptic neurons (post_to_pre, default) or post-synaptic neurons (pre_to_post). Note that affects only the C++ data structures.
    * `only_int_idx_type`: if set to True (default) only signed integers are used to store pre-/post-synaptic ranks which was default until 4.7.If set to False, the index type used in a single projection is selected based on the size of the corresponding populations.
    * `visible_cores`: allows a fine-grained control which cores are useable for the created threads (default = [] for no limitation). It can be used to limit created openMP threads to a physical socket.

    The following parameters are mainly for debugging and profiling, and should be ignored by most users:

    * `verbose`: shows details about compilation process on console (by default False). Additional some information of the network construction will be shown.
    * `suppress_warnings`: if True, warnings (e. g. from the mathematical parser) are suppressed.
    * `show_time`: if True, initialization times are shown. Attention: verbose should be set to True additionally.
    * `disable_shared_library_time_offset`: by default False. If set to True, the shared library generated by ANNarchy will not be extended by time offset.

    **Note:**

    This function should be used before any other functions of ANNarchy (including importing a network definition), right after `import ANNarchy`:

    ```python
    import ANNarchy as ann
    ann.setup(dt=1.0, method='midpoint', num_threads=2)
    ```

    """
    if len(NetworkManager().get_network(0).get_populations()) > 0:
        if 'dt' in keyValueArgs:
            Messages._warning('setup(): populations or projections have already been created at the global level. Changing dt now might lead to strange behaviors with the synaptic delays (internally generated in steps, not ms)...')
        if 'precision' in keyValueArgs:
            Messages._warning('setup(): populations or projections have already been created at the global level. Changing precision now might lead to strange behaviors...')

    config_manager = ConfigManager()

    for key in keyValueArgs:
        # sanity check: filter out performance flags
        if key in ConfigManager()._performance_related_config_keys:
            Messages._error("Performance related flags can not be configured by setup()")

        if key in ConfigManager().keys():
            ConfigManager().set(key, keyValueArgs[key], net_id=0)
        else:
            Messages._warning('setup(): unknown key:', key)

        if key == 'seed': # also seed numpy, but this is the old way
            np.random.seed(keyValueArgs[key])

        if key == 'sparse_matrix_format':
            # check if this is a supported format
            if keyValueArgs[key] not in ["lil", "csr", "csr_vector", "csr_scalar", "dense", "ell", "ellr", "sell", "coo", "bsr", "hyb", "auto"]:
                Messages._error("The value", keyValueArgs[key], "provided to sparse_matrix_format is not valid.")

def _optimization_flags(**keyValueArgs):
    """
    In particular the ANNarchy 4.7.x releases added various optional arguments to control the code generation. Please take in mind, that these flags might not being tested thoroughly on all features available in ANNarchy. They are intended for experimental features or performance analysis.

    * disable_parallel_rng: determines if random numbers drawn from distributions are generated from a single source (default: True).  If this flag is set to true only one RNG source is used und the values are drawn by one thread which reduces parallel performance (this is the behavior of all ANNarchy versions prior to 4.7). If set to false a seed sequence is generated to allow usage of one RNG per thread. Please note, that this lag won't effect the GPUs which draw from multiple sources anyways.
    * use_seed_seq: If parallel RNGs are used the single generators need to be initialized. By default (use_seed_seq == True) we use the STL seed sequence to generate a list of seeds from the given master seed (*seed* argument). If set to False, we use an improved version of the sequence generator proposed by M.E. O'Neill (https://www.pcg-random.org/posts/simple-portable-cpp-seed-entropy.html)
    * use_cpp_connectors:   For some of the default connectivity methods of ANNarchy we offer a CPP-side construction of the pattern to improve the initialization time (default=False). For maximum performance the disable_parallel_rng should be set to False to allow a parallel construction of the pattern.
    * disable_split_matrix: determines if projections can use thread-local allocation. If set to *True* (default) no thread local allocation is allowed. This equals the behavior of ANNarchy until 4.7. If set to *False* the code generator can use sliced versions if they are available.
    * disable_SIMD_SpMV: determines if the hand-written implementation is used (by default False) if the current hardware platform and used sparse matrix format does support the vectorization). Disabling is intended for performance analysis.

    * disable_SIMD_Eq: this flags disables auto-vectorization and openMP simd (by default False). Disabling is intended for performance analysis.

    **Note:**

    This function should be used only for special purposes therefore its not publicly available.

    ```python
    from ANNarchy import *  # will not work
    from ANNarchy.core.Global import _optimization_flags
    _optimization_flags(disable_parallel_rng=False)
    ```

    """
    for key in keyValueArgs:
        # Sanity check: valid key?
        if key not in ConfigManager().keys():
            Messages._warning('_optimization_flags() received unknown key:', key)
            continue

        if key not in ConfigManager()._performance_related_config_keys:
            Messages._warning(f"The key '{key}' does not belong to the performance related keys.")
            continue

        # Update global config
        _update_global_config(key, keyValueArgs[key])

        # Warning: use_cpp_connectors and disable_parallel_rng should be both activated
        if key == "use_cpp_connectors":
            if get_global_config(key) == True:
                Messages._warning("use_cpp_connectors is an experimental feature, we greatly appreciate bug reports.")

                # check if the key is in the update list
                if 'disable_parallel_rng' in keyValueArgs.keys():
                    if keyValueArgs['disable_parallel_rng']:
                        Messages._warning("If 'use_cpp_connectors' is enabled, the 'disable_parallel_rng' flag should be disabled for maximum efficiency.")
                # is it enabled by default?
                elif get_global_config('disable_parallel_rng'):
                    Messages._warning("If 'use_cpp_connectors' is enabled, the 'disable_parallel_rng' flag should be disabled for maximum efficiency.")
                # no conflict
                else:
                    pass

#############################################
# Globally available functions (internal)
#############################################
def get_global_config(key: str) -> Union[str,float,bool]:
    """
    Returns a global configuration.
    """
    return ConfigManager().get(key, net_id=0)

def _update_global_config(key: str, value: Union[str,float,bool]) -> None:
    """
    Updates a global configuration flag.

    Note: this function is intended for internal use.
          As user, please refer to *setup()* method.
    """
    return ConfigManager().set(key, value, net_id=0)

def _check_paradigm(paradigm, net_id=0):
    """
    Returns True when the provided paradigm is currently used.

    Possible values:

    1. "openmp"
    2. "cuda"
    """
    try:
        return paradigm == ConfigManager().get('paradigm', net_id)
    except KeyError:
        Messages._error("Unknown paradigm")

def _check_precision(precision, net_id=0):
    """
    Returns True when the provided precision is currently used.

    Possible values:

    1. "float"
    2. "double"
    """
    try:
        return precision == ConfigManager().get('precision', net_id)
    except KeyError:
        Messages._error("Unknown precision")



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

attribute_decl = {
    'local':
"""
    // Local %(attr_type)s %(name)s
    std::vector< %(type)s > %(name)s;
""",
    'semiglobal':
"""
    // Semiglobal %(attr_type)s %(name)s
    std::vector< %(type)s >  %(name)s ;
""",
    'global':
"""
    // Global %(attr_type)s %(name)s
    %(type)s  %(name)s ;
"""
}

attribute_cpp_init = {
    'local':
"""
        // Local %(attr_type)s %(name)s
        %(name)s = init_matrix_variable<%(type)s>(static_cast<%(type)s>(%(init)s));
        if(%(name)s.empty())
            return false;
""",
    'semiglobal':
"""
        // Semiglobal %(attr_type)s %(name)s
        %(name)s = init_vector_variable<%(type)s>(static_cast<%(type)s>(%(init)s));
""",
    'global':
"""
        // Global %(attr_type)s %(name)s
        %(name)s = %(init)s;
"""
}

attribute_cpp_size = {
    'local': """
        // Local %(attr_type)s %(name)s
        size_in_bytes += sizeof(std::vector<%(ctype)s>);
        size_in_bytes += sizeof(%(ctype)s) * %(name)s.capacity();
""",
    'semiglobal': """
        // Semiglobal %(attr_type)s %(name)s
        size_in_bytes += sizeof(std::vector<%(ctype)s>);
        size_in_bytes += sizeof(%(ctype)s) * %(name)s.capacity();
""",
    'global': """
        // Global %(attr_type)s %(name)s
        size_in_bytes += sizeof(%(ctype)s);
"""
}

attribute_cpp_delete = {
    'local': """
        // %(name)s
        %(name)s.clear();
        %(name)s.shrink_to_fit();
""",
    'semiglobal': """
        // %(name)s
        %(name)s.clear();
        %(name)s.shrink_to_fit();
""",
    'global': ""
}

delay = {
    # A single value for all synapses
    'uniform': {
        'declare': """
    // Uniform delay
    int delay;

    int get_delay() { return delay; }
    int get_dendrite_delay(int idx) { return delay; }
    void set_delay(int delay) { this->delay = delay; }
""",
        'init': """
    delay = delays[0][0];
"""
    },
    # An individual value for each synapse
    'nonuniform_rate_coded': None,
    # An individual value for each synapse and a
    # buffer for spike events
    'nonuniform_spiking': None
}

spiking_summation_fixed_delay = """// Event-based summation
if (_transmission && %(post_prefix)s_active){

    for (%(idx_type)s rk_post = 0; rk_post < num_rows(); rk_post++) {
        // Iterate over all spiking neurons
        for (auto it = %(pre_prefix)sspiked.cbegin(); it != %(pre_prefix)sspiked.cend(); it++) {
            %(idx_type)s rk_pre = (*it);
            if ((rk_pre < this->low_column_rank_) || (rk_pre >= this->high_column_rank_))
                continue;

            %(size_type)s j = rk_post*this->num_columns_ + *it;

            %(g_target)s

            if (mask_[j]) {
                %(event_driven)s
                %(pre_event)s
            }
        }
    }
} // active
"""

conn_templates = {
    # accessors
    'attribute_decl': attribute_decl,
    'attribute_cpp_init': attribute_cpp_init,
    'attribute_cpp_size': attribute_cpp_size,
    'attribute_cpp_delete': attribute_cpp_delete,
    'delay': delay,

    #operations
    'rate_coded_sum': None,
    'vectorized_default_psp': None,
    'spiking_sum_fixed_delay': spiking_summation_fixed_delay,
    'update_variables': None,
}

conn_ids = {
    'local_index': '[j]',
    'semiglobal_index': '[rk_post]',
    'global_index': '',
    'post_index': '[rk_post]',
    'pre_index': '[rk_pre]',
    'delay_u' : '[delay-1]' # uniform delay
}